@@ -12,7 +12,7 @@ svc = bentoml.Service("iris_classifier", runners=[iris_clf_runner])
     input=NumpyNdarray.from_sample(
         np.array([[4.9, 3.0, 1.4, 0.2]], dtype=np.double), enforce_shape=False
     ),
-    output=NumpyNdarray(),
+    output=NumpyNdarray.from_sample(np.array([0.0], dtype=np.double)),
 )
 async def classify(input_series: np.ndarray) -> np.ndarray:
     return await iris_clf_runner.predict.async_run(input_series)
