@@ -16,23 +16,23 @@ import yaml
 import psutil
 import fs.copy
 from pathspec import PathSpec
-from pip_requirements_parser import RequirementsFile
 
-from .gen import generate_dockerfile
 from ..utils import bentoml_cattr
 from ..utils import resolve_user_filepath
 from ..utils import copy_file_to_fs_folder
-from .docker import DistroSpec
-from .docker import get_supported_spec
-from .docker import SUPPORTED_CUDA_VERSIONS
-from .docker import DOCKER_SUPPORTED_DISTROS
-from .docker import ALLOWED_CUDA_VERSION_ARGS
-from .docker import SUPPORTED_PYTHON_VERSIONS
+from ..container import generate_containerfile
 from ...exceptions import InvalidArgument
 from ...exceptions import BentoMLException
 from ..utils.dotenv import parse_dotenv
 from ..configuration import CLEAN_BENTOML_VERSION
+from ..container.generate import BENTO_PATH
 from .build_dev_bentoml_whl import build_bentoml_editable_wheel
+from ..container.frontend.dockerfile import DistroSpec
+from ..container.frontend.dockerfile import get_supported_spec
+from ..container.frontend.dockerfile import SUPPORTED_CUDA_VERSIONS
+from ..container.frontend.dockerfile import ALLOWED_CUDA_VERSION_ARGS
+from ..container.frontend.dockerfile import SUPPORTED_PYTHON_VERSIONS
+from ..container.frontend.dockerfile import CONTAINER_SUPPORTED_DISTROS
 
 if TYPE_CHECKING:
     from attr import Attribute
@@ -43,7 +43,7 @@ logger = logging.getLogger(__name__)
 
 # Docker defaults
 DEFAULT_CUDA_VERSION = "11.6.2"
-DEFAULT_DOCKER_DISTRO = "debian"
+DEFAULT_CONTAINER_DISTRO = "debian"
 
 CONDA_ENV_YAML_FILE_NAME = "environment.yml"
 
@@ -111,12 +111,23 @@ def _convert_env(
     if isinstance(env, list):
         env_dict: dict[str, str | None] = {}
         for envvar in env:
-            match = re.match(r"^(\w+)=([\w-]+)$", envvar)
-            if not match:
+            if not re.match(r"^(\w+)=(?:[\.\w\-,\/]+)$", envvar):
                 raise BentoMLException(
                     "All value in `env` list must follow format ENV=VALUE"
                 )
-            env_key, env_value = match.groups()
+            env_key, _, env_value = envvar.partition("=")
+            if os.path.isfile(env_value):
+                bento_env_path = BENTO_PATH + os.path.abspath(
+                    os.path.expanduser(os.path.expandvars(env_value))
+                )
+                logger.info(
+                    "'%s' sets to '%s', which is a file. Make sure to mount this file as a persistent volume to the container when using 'run' command: 'docker run -v %s:%s ...'",
+                    env_key,
+                    env_value,
+                    env_value,
+                    bento_env_path,
+                )
+                env_value = bento_env_path
             env_dict[env_key] = env_value
         return env_dict
 
@@ -131,7 +142,6 @@ def _convert_env(
 
 @attr.frozen
 class DockerOptions:
-
     # For validating user defined bentofile.yaml.
     __forbid_extra_keys__ = True
     # always omit config values in case of default values got changed in future BentoML releases
@@ -140,7 +150,7 @@ class DockerOptions:
     distro: t.Optional[str] = attr.field(
         default=None,
         validator=attr.validators.optional(
-            attr.validators.in_(DOCKER_SUPPORTED_DISTROS)
+            attr.validators.in_(CONTAINER_SUPPORTED_DISTROS)
         ),
     )
     python_version: t.Optional[str] = attr.field(
@@ -206,29 +216,34 @@ class DockerOptions:
 
         if self.base_image is None:
             if self.distro is None:
-                defaults["distro"] = DEFAULT_DOCKER_DISTRO
+                defaults["distro"] = DEFAULT_CONTAINER_DISTRO
             if self.python_version is None:
                 python_version = f"{version_info.major}.{version_info.minor}"
                 defaults["python_version"] = python_version
 
         return attr.evolve(self, **defaults)
 
-    def write_to_bento(
-        self, bento_fs: FS, build_ctx: str, conda_options: CondaOptions
-    ) -> None:
-        use_conda = not conda_options.is_empty()
-
+    def write_to_bento(self, bento_fs: FS, build_ctx: str, conda: CondaOptions) -> None:
         docker_folder = fs.path.combine("env", "docker")
         bento_fs.makedirs(docker_folder, recreate=True)
-        dockerfile = fs.path.combine(docker_folder, "Dockerfile")
-
-        with bento_fs.open(dockerfile, "w") as dockerfile:
-            dockerfile.write(
-                generate_dockerfile(self, build_ctx=build_ctx, use_conda=use_conda)
-            )
+        dockerfile_path = fs.path.combine(docker_folder, "Dockerfile")
 
+        # NOTE that by default the generated Dockerfile won't have BuildKit syntax.
+        # By default, BentoML containerization will use BuildKit. To opt-out specify DOCKER_BUILDKIT=0
+        bento_fs.writetext(
+            dockerfile_path,
+            generate_containerfile(
+                self, build_ctx, conda=conda, bento_fs=bento_fs, enable_buildkit=False
+            ),
+        )
         copy_file_to_fs_folder(
-            fs.path.join(os.path.dirname(__file__), "docker", "entrypoint.sh"),
+            fs.path.join(
+                os.path.dirname(os.path.dirname(__file__)),
+                "container",
+                "frontend",
+                "dockerfile",
+                "entrypoint.sh",
+            ),
             bento_fs,
             docker_folder,
         )
@@ -249,10 +264,23 @@ class DockerOptions:
                 setup_script, bento_fs, docker_folder, "setup_script"
             )
 
+        # If dockerfile_template is provided, then we copy it to the Bento
+        # This template then can be used later to containerize.
+        if self.dockerfile_template is not None:
+            copy_file_to_fs_folder(
+                resolve_user_filepath(self.dockerfile_template, build_ctx),
+                bento_fs,
+                docker_folder,
+                "Dockerfile.template",
+            )
+
+    def to_dict(self) -> dict[str, t.Any]:
+        return bentoml_cattr.unstructure(self)
+
 
 if TYPE_CHECKING:
-    CondaPipType = t.Dict[t.Literal["pip"], t.List[str]]
-    DependencyType = t.List[t.Union[str, CondaPipType]]
+    CondaPipType = dict[t.Literal["pip"], list[str]]
+    DependencyType = list[str | CondaPipType]
 else:
     DependencyType = list
 
@@ -286,13 +314,13 @@ def conda_dependencies_validator(
 
 if TYPE_CHECKING:
     ListStr: t.TypeAlias = list[str]
+    CondaYamlDict = dict[str, DependencyType | list[str]]
 else:
     ListStr = list
 
 
 @attr.frozen
 class CondaOptions:
-
     # User shouldn't add new fields under yaml file.
     __forbid_extra_keys__ = True
     # no need to omit since BentoML has already handled the default values.
@@ -350,26 +378,48 @@ class CondaOptions:
                 dst_filename=CONDA_ENV_YAML_FILE_NAME,
             )
         else:
-            deps_list = [] if self.dependencies is None else self.dependencies
+            deps_list: DependencyType = []
+            if self.dependencies is not None:
+                deps_list.extend(self.dependencies)
             if self.pip is not None:
                 if any(isinstance(x, dict) for x in deps_list):
                     raise BentoMLException(
                         "Cannot not have both 'conda.dependencies.pip' and 'conda.pip'"
                     )
-                deps_list.append(dict(pip=self.pip))  # type: ignore
+                deps_list.append({"pip": self.pip})
 
             if not deps_list:
                 return
 
-            yaml_content: dict[str, DependencyType | list[str] | None] = dict(
-                dependencies=deps_list
-            )
+            yaml_content: CondaYamlDict = {"dependencies": deps_list}
+            assert self.channels is not None
             yaml_content["channels"] = self.channels
             with bento_fs.open(
                 fs.path.combine(conda_folder, CONDA_ENV_YAML_FILE_NAME), "w"
             ) as f:
                 yaml.dump(yaml_content, f)
 
+    def get_python_version(self, bento_fs: FS) -> str | None:
+        # Get the python version from given environment.yml file
+
+        environment_yml = bento_fs.getsyspath(
+            fs.path.join(
+                "env",
+                "conda",
+                CONDA_ENV_YAML_FILE_NAME,
+            )
+        )
+        if os.path.exists(environment_yml):
+            with open(environment_yml, "r") as f:
+                for line in f:
+                    match = re.search(r"(?:python=)(\d+.\d+)$", line)
+                    if match:
+                        return match.group().split("=")[-1]
+            logger.debug(
+                "No python version is specified under '%s'. Using the Python options specified under 'docker'.",
+                environment_yml,
+            )
+
     def with_defaults(self) -> CondaOptions:
         # Convert from user provided options to actual build options with default values
         defaults: dict[str, t.Any] = {}
@@ -396,7 +446,6 @@ class CondaOptions:
 
 @attr.frozen
 class PythonOptions:
-
     # User shouldn't add new fields under yaml file.
     __forbid_extra_keys__ = True
     # no need to omit since BentoML has already handled the default values.
@@ -472,7 +521,7 @@ class PythonOptions:
         build_bentoml_editable_wheel(bento_fs.getsyspath(wheels_folder))
 
         # Move over required wheel files
-        if self.wheels is not None:
+        if self.wheels:
             bento_fs.makedirs(wheels_folder, recreate=True)
             for whl_file in self.wheels:  # pylint: disable=not-an-iterable
                 whl_file = resolve_user_filepath(whl_file, build_ctx)
@@ -555,6 +604,8 @@ fi
             f.write(install_sh)
 
         if self.requirements_txt is not None:
+            from pip_requirements_parser import RequirementsFile
+
             requirements_txt = RequirementsFile.from_file(
                 resolve_user_filepath(self.requirements_txt, build_ctx),
                 include_nested=True,
@@ -601,6 +652,7 @@ fi
                     "--allow-unsafe",
                     "--no-header",
                     f"--output-file={pip_compile_out}",
+                    "--resolver=backtracking",
                 ]
             )
             logger.info("Locking PyPI package versions.")
@@ -670,6 +722,7 @@ class BentoBuildConfig:
     __omit_if_default__ = False
 
     service: str
+    name: t.Optional[str] = None
     description: t.Optional[str] = None
     labels: t.Optional[t.Dict[str, t.Any]] = None
     include: t.Optional[t.List[str]] = None
@@ -688,7 +741,6 @@ class BentoBuildConfig:
     )
 
     if TYPE_CHECKING:
-
         # NOTE: This is to ensure that BentoBuildConfig __init__
         # satisfies type checker. docker, python, and conda accepts
         # dict[str, t.Any] since our converter will handle the conversion.
@@ -697,6 +749,7 @@ class BentoBuildConfig:
         def __init__(
             self,
             service: str,
+            name: str | None = ...,
             description: str | None = ...,
             labels: dict[str, t.Any] | None = ...,
             include: list[str] | None = ...,
@@ -712,7 +765,7 @@ class BentoBuildConfig:
         use_cuda = self.docker.cuda_version is not None
 
         if use_cuda and use_conda:
-            raise BentoMLException(
+            logger.warning(
                 "BentoML does not support using both conda dependencies and setting a CUDA version for GPU. If you need both conda and CUDA, use a custom base image or create a dockerfile_template, see https://docs.bentoml.org/en/latest/concepts/bento.html#custom-base-image-advanced"
             )
 
@@ -722,31 +775,26 @@ class BentoBuildConfig:
                     f"{self.docker.distro} does not supports conda. BentoML will only support conda with the following distros: {get_supported_spec('miniconda')}."
                 )
 
-            _spec = DistroSpec.from_distro(
-                self.docker.distro, cuda=use_cuda, conda=use_conda
-            )
+            spec = DistroSpec.from_options(self.docker, self.conda)
+            if (
+                self.docker.python_version is not None
+                and self.docker.python_version not in spec.supported_python_versions
+            ):
+                raise BentoMLException(
+                    f"{self.docker.python_version} is not supported for {self.docker.distro}. Supported python versions are: {', '.join(spec.supported_python_versions)}."
+                )
 
-            if _spec is not None:
-                if self.docker.python_version is not None:
-                    if (
-                        self.docker.python_version
-                        not in _spec.supported_python_versions
-                    ):
-                        raise BentoMLException(
-                            f"{self.docker.python_version} is not supported for {self.docker.distro}. Supported python versions are: {', '.join(_spec.supported_python_versions)}."
-                        )
-
-                if self.docker.cuda_version is not None:
-                    if _spec.supported_cuda_versions is None:
-                        raise BentoMLException(
-                            f"{self.docker.distro} does not support CUDA, while 'docker.cuda_version={self.docker.cuda_version}' is provided."
-                        )
-                    if self.docker.cuda_version != "default" and (
-                        self.docker.cuda_version not in _spec.supported_cuda_versions
-                    ):
-                        raise BentoMLException(
-                            f"{self.docker.cuda_version} is not supported for {self.docker.distro}. Supported cuda versions are: {', '.join(_spec.supported_cuda_versions)}."
-                        )
+            if self.docker.cuda_version is not None:
+                if spec.supported_cuda_versions is None:
+                    raise BentoMLException(
+                        f"{self.docker.distro} does not support CUDA, while 'docker.cuda_version={self.docker.cuda_version}' is provided."
+                    )
+                if self.docker.cuda_version != "default" and (
+                    self.docker.cuda_version not in spec.supported_cuda_versions
+                ):
+                    raise BentoMLException(
+                        f"{self.docker.cuda_version} is not supported for {self.docker.distro}. Supported cuda versions are: {', '.join(spec.supported_cuda_versions)}."
+                    )
 
     def with_defaults(self) -> FilledBentoBuildConfig:
         """
@@ -758,6 +806,7 @@ class BentoBuildConfig:
         """
         return FilledBentoBuildConfig(
             self.service,
+            self.name,
             self.description,
             {} if self.labels is None else self.labels,
             ["*"] if self.include is None else self.include,
@@ -837,6 +886,7 @@ class BentoPathSpec:
 
 class FilledBentoBuildConfig(BentoBuildConfig):
     service: str
+    name: t.Optional[str]
     description: t.Optional[str]
     labels: t.Dict[str, t.Any]
     include: t.List[str]
