@@ -3,14 +3,15 @@ from __future__ import annotations
 import typing as t
 from abc import ABC
 from abc import abstractmethod
-from typing import TYPE_CHECKING
+from functools import update_wrapper
 
 from ...exceptions import InvalidArgument
 
-if TYPE_CHECKING:
+if t.TYPE_CHECKING:
     from types import UnionType
 
-    from typing_extensions import Self
+    import pyarrow
+    import pyspark.sql.types
     from starlette.requests import Request
     from starlette.responses import Response
 
@@ -19,7 +20,6 @@ if TYPE_CHECKING:
     from ..types import LazyType
     from ..context import InferenceApiContext as Context
     from ..service.openapi.specification import Schema
-    from ..service.openapi.specification import MediaType
     from ..service.openapi.specification import Reference
 
     InputType = (
@@ -28,7 +28,8 @@ if TYPE_CHECKING:
         | LazyType[t.Any]
         | dict[str, t.Type[t.Any] | UnionType | LazyType[t.Any]]
     )
-    OpenAPIResponse = dict[str, str | dict[str, MediaType] | dict[str, t.Any]]
+    OpenAPIResponse = dict[str, str | dict[str, t.Any]]
+    F = t.Callable[..., t.Any]
 
 
 IO_DESCRIPTOR_REGISTRY: dict[str, type[IODescriptor[t.Any]]] = {}
@@ -36,9 +37,13 @@ IO_DESCRIPTOR_REGISTRY: dict[str, type[IODescriptor[t.Any]]] = {}
 IOType = t.TypeVar("IOType")
 
 
-def from_spec(spec: dict[str, str]) -> IODescriptor[t.Any]:
+def from_spec(spec: dict[str, t.Any]) -> IODescriptor[t.Any]:
     if "id" not in spec:
         raise InvalidArgument(f"IO descriptor spec ({spec}) missing ID.")
+
+    if spec["id"] is None:
+        raise InvalidArgument("No IO descriptor spec found.")
+
     return IO_DESCRIPTOR_REGISTRY[spec["id"]].from_spec(spec)
 
 
@@ -71,20 +76,31 @@ class IODescriptor(ABC, _OpenAPIMeta, t.Generic[IOType]):
     endpoint IO descriptor types in BentoServer.
     """
 
-    __slots__ = ("_args", "_kwargs", "_proto_fields", "_mime_type", "descriptor_id")
+    if t.TYPE_CHECKING:
+        # Populated by subclasses. Makes pyright happy.
+        def __init__(self, **kwargs: t.Any) -> None:
+            ...
 
-    HTTP_METHODS = ["POST"]
+    __slots__ = (
+        "_mime_type",
+        "proto_fields",
+        "descriptor_id",
+        "from_sample",
+        "_sample",
+    )
 
-    descriptor_id: str | None
+    HTTP_METHODS = ("POST",)
 
     _mime_type: str
-    _rpc_content_type: str = "application/grpc"
-    _proto_fields: tuple[ProtoField]
-    _sample: IOType | None = None
-    _args: t.Sequence[t.Any]
-    _kwargs: dict[str, t.Any]
+    descriptor_id: str | None
+    proto_fields: tuple[ProtoField]
 
-    def __init_subclass__(cls, *, descriptor_id: str | None = None):
+    def __init_subclass__(
+        cls,
+        *,
+        descriptor_id: str | None = None,
+        proto_fields: tuple[ProtoField] | None = None,
+    ):
         if descriptor_id is not None:
             if descriptor_id in IO_DESCRIPTOR_REGISTRY:
                 raise ValueError(
@@ -92,11 +108,22 @@ class IODescriptor(ABC, _OpenAPIMeta, t.Generic[IOType]):
                 )
             IO_DESCRIPTOR_REGISTRY[descriptor_id] = cls
         cls.descriptor_id = descriptor_id
+        cls.proto_fields = proto_fields or t.cast("tuple[ProtoField]", ())
 
-    if TYPE_CHECKING:
+        cls._sample: IOType | None = None
 
-        def __init__(self, **kwargs: t.Any) -> None:
-            ...
+        def _() -> F:
+            def impl(
+                cl_: type[IODescriptor[t.Any]], sample: t.Any, **kwargs: t.Any
+            ) -> IODescriptor[t.Any]:
+                klass = cl_(**kwargs)
+                klass.sample = klass._from_sample(sample)
+                return klass
+
+            impl.__doc__ = cls._from_sample.__doc__
+            return update_wrapper(impl, cls._from_sample)
+
+        cls.from_sample = classmethod(_())
 
     def __repr__(self) -> str:
         return self.__class__.__qualname__
@@ -109,27 +136,29 @@ class IODescriptor(ABC, _OpenAPIMeta, t.Generic[IOType]):
     def sample(self, value: IOType) -> None:
         self._sample = value
 
-    @classmethod
-    def from_sample(cls, sample: IOType | t.Any, **kwargs: t.Any) -> Self:
-        klass = cls(**kwargs)
-        klass.sample = klass._from_sample(sample)
-        return klass
-
     @abstractmethod
     def _from_sample(self, sample: t.Any) -> IOType:
+        """
+        Creates a new instance of the IO Descriptor from given sample.
+
+        Args:
+            sample: The sample to create the instance from.
+            **kwargs: Additional keyword arguments to pass to the constructor.
+
+        Returns:
+            An instance of the IODescriptor.
+        """
         raise NotImplementedError
 
     @property
     def mime_type(self) -> str:
         return self._mime_type
 
-    @abstractmethod
-    def to_spec(self) -> dict[str, t.Any]:
-        raise NotImplementedError
+    def to_spec(self) -> dict[str, t.Any] | None:
+        return None
 
     @classmethod
-    @abstractmethod
-    def from_spec(cls, spec: dict[str, t.Any]) -> Self:
+    def from_spec(cls, spec: dict[str, t.Any]) -> t.Self:
         raise NotImplementedError
 
     @abstractmethod
@@ -153,3 +182,18 @@ class IODescriptor(ABC, _OpenAPIMeta, t.Generic[IOType]):
     @abstractmethod
     async def to_proto(self, obj: IOType) -> t.Any:
         raise NotImplementedError
+
+    def from_arrow(self, batch: pyarrow.RecordBatch) -> IOType:
+        raise NotImplementedError(
+            "This IO descriptor does not currently support batch inference."
+        )
+
+    def to_arrow(self, obj: IOType) -> pyarrow.RecordBatch:
+        raise NotImplementedError(
+            "This IO descriptor does not currently support batch inference."
+        )
+
+    def spark_schema(self) -> pyspark.sql.types.StructType:
+        raise NotImplementedError(
+            "This IO descriptor does not currently support batch inference."
+        )
