@@ -3,7 +3,6 @@ from __future__ import annotations
 import io
 import typing as t
 import functools
-from typing import TYPE_CHECKING
 from urllib.parse import quote
 
 from starlette.requests import Request
@@ -20,37 +19,36 @@ from ...exceptions import BadInput
 from ...exceptions import InvalidArgument
 from ...exceptions import InternalServerError
 from ...exceptions import MissingDependencyException
+from ...grpc.utils import import_generated_stubs
 from ..service.openapi import SUCCESS_DESCRIPTION
 from ..service.openapi.specification import Schema
 from ..service.openapi.specification import MediaType
 
 PIL_EXC_MSG = "'Pillow' is required to use the Image IO descriptor. Install with 'pip install bentoml[io-image]'."
 
-if TYPE_CHECKING:
+if t.TYPE_CHECKING:
     from types import UnionType
 
     import PIL
     import PIL.Image
-    from typing_extensions import Self
-
-    from bentoml.grpc.v1alpha1 import service_pb2 as pb
 
     from .. import external_typing as ext
     from .base import OpenAPIResponse
     from ..context import InferenceApiContext as Context
+    from ...grpc.v1 import service_pb2 as pb
+    from ...grpc.v1alpha1 import service_pb2 as pb_v1alpha1
 
     _Mode = t.Literal[
         "1", "CMYK", "F", "HSV", "I", "L", "LAB", "P", "RGB", "RGBA", "RGBX", "YCbCr"
     ]
 else:
-    from bentoml.grpc.utils import import_generated_stubs
-
     # NOTE: pillow-simd only benefits users who want to do preprocessing
     # TODO: add options for users to choose between simd and native mode
     PIL = LazyLoader("PIL", globals(), "PIL", exc_msg=PIL_EXC_MSG)
     PIL.Image = LazyLoader("PIL.Image", globals(), "PIL.Image", exc_msg=PIL_EXC_MSG)
 
-    pb, _ = import_generated_stubs()
+    pb, _ = import_generated_stubs("v1")
+    pb_v1alpha1, _ = import_generated_stubs("v1alpha1")
 
 # NOTES: we will keep type in quotation to avoid backward compatibility
 #  with numpy < 1.20, since we will use the latest stubs from the main branch of numpy.
@@ -80,7 +78,9 @@ def initialize_pillow():
     READABLE_MIMES = {k for k, v in MIME_EXT_MAPPING.items() if v not in PIL_WRITE_ONLY_FORMATS}  # type: ignore (lazy constant)
 
 
-class Image(IODescriptor[ImageType], descriptor_id="bentoml.io.Image"):
+class Image(
+    IODescriptor[ImageType], descriptor_id="bentoml.io.Image", proto_fields=("file",)
+):
     """
     :obj:`Image` defines API specification for the inputs/outputs of a Service, where either
     inputs will be converted to or outputs will be converted from images as specified
@@ -167,8 +167,6 @@ class Image(IODescriptor[ImageType], descriptor_id="bentoml.io.Image"):
         :obj:`Image`: IO Descriptor that either a :code:`PIL.Image.Image` or a :code:`np.ndarray` representing an image.
     """
 
-    _proto_fields = ("file",)
-
     def __init__(
         self,
         pilmode: _Mode | None = DEFAULT_PIL_MODE,
@@ -211,6 +209,42 @@ class Image(IODescriptor[ImageType], descriptor_id="bentoml.io.Image"):
         self._format: str = MIME_EXT_MAPPING[self._mime_type]
 
     def _from_sample(self, sample: ImageType | str) -> ImageType:
+        """
+        Create a :class:`~bentoml._internal.io_descriptors.image.Image` IO Descriptor from given inputs.
+
+        Args:
+            sample: Given File-like object, or a path to a file.
+            pilmode: Optional color mode for PIL. Default to ``RGB``.
+            mime_type: The MIME type of the file type that this descriptor should return.
+                       If not specified, then ``from_sample`` will try to infer the MIME type
+                       from file extension.
+            allowed_mime_types: An optional list of MIME types to restrict input to.
+
+        Returns:
+            :class:`~bentoml._internal.io_descriptors.image.Image`: IODescriptor from given users inputs.
+
+        Example:
+
+        .. code-block:: python
+           :caption: `service.py`
+
+           from __future__ import annotations
+
+           import bentoml
+           from typing import Any
+           from bentoml.io import Image
+           import numpy as np
+
+           input_spec = Image.from_sample("/path/to/image.jpg")
+           @svc.api(input=input_spec, output=Image())
+           async def predict(input: t.IO[t.Any]) -> t.IO[t.Any]:
+               return await runner.async_run(input)
+
+        Raises:
+            :class:`InvalidArgument`: If the given sample is not a valid image type.
+            :class:`MissingDependencyException`: If ``filetype`` is not installed.
+            :class:`BadInput`: Given sample from file can't be parsed with Pillow.
+        """
         try:
             from filetype.match import image_match
         except ModuleNotFoundError:
@@ -245,7 +279,7 @@ class Image(IODescriptor[ImageType], descriptor_id="bentoml.io.Image"):
         }
 
     @classmethod
-    def from_spec(cls, spec: dict[str, t.Any]) -> Self:
+    def from_spec(cls, spec: dict[str, t.Any]) -> t.Self:
         if "args" not in spec:
             raise InvalidArgument(f"Missing args key in Image spec: {spec}")
 
@@ -387,15 +421,13 @@ class Image(IODescriptor[ImageType], descriptor_id="bentoml.io.Image"):
                 headers={"content-disposition": content_disposition},
             )
 
-    async def from_proto(self, field: pb.File | bytes) -> ImageType:
-        from bentoml.grpc.utils import filetype_pb_to_mimetype_map
-
-        mapping = filetype_pb_to_mimetype_map()
-        # check if the request message has the correct field
+    async def from_proto(self, field: pb.File | pb_v1alpha1.File | bytes) -> ImageType:
         if isinstance(field, bytes):
             content = field
-        else:
-            assert isinstance(field, pb.File)
+        elif isinstance(field, pb_v1alpha1.File):
+            from .file import filetype_pb_to_mimetype_map
+
+            mapping = filetype_pb_to_mimetype_map()
             if field.kind:
                 try:
                     mime_type = mapping[field.kind]
@@ -405,16 +437,32 @@ class Image(IODescriptor[ImageType], descriptor_id="bentoml.io.Image"):
                         )
                 except KeyError:
                     raise BadInput(
-                        f"{field.kind} is not a valid File kind. Accepted file kind: {[names for names,_ in pb.File.FileType.items()]}",
+                        f"{field.kind} is not a valid File kind. Accepted file kind: {[names for names,_ in pb_v1alpha1.File.FileType.items()]}",
                     ) from None
+            if not field.content:
+                raise BadInput("Content is empty!") from None
+            return PIL.Image.open(io.BytesIO(field.content))
+        else:
+            assert isinstance(field, pb.File)
+            if field.kind and field.kind != self._mime_type:
+                raise BadInput(
+                    f"MIME type from 'kind' is '{field.kind}', while '{self!r}' is expecting '{self._mime_type}'",
+                )
             content = field.content
             if not content:
                 raise BadInput("Content is empty!") from None
 
         return PIL.Image.open(io.BytesIO(content))
 
-    async def to_proto(self, obj: ImageType) -> pb.File:
-        from bentoml.grpc.utils import mimetype_to_filetype_pb_map
+    async def to_proto_v1alpha1(self, obj: ImageType) -> pb_v1alpha1.File:
+        from .file import mimetype_to_filetype_pb_map
+
+        try:
+            kind = mimetype_to_filetype_pb_map()[self._mime_type]
+        except KeyError:
+            raise BadInput(
+                f"{self._mime_type} doesn't have a corresponding File 'kind'"
+            ) from None
 
         if LazyType["ext.NpNDArray"]("numpy.ndarray").isinstance(obj):
             image = PIL.Image.fromarray(obj, mode=self._pilmode)
@@ -427,11 +475,18 @@ class Image(IODescriptor[ImageType], descriptor_id="bentoml.io.Image"):
         ret = io.BytesIO()
         image.save(ret, format=self._format)
 
-        try:
-            kind = mimetype_to_filetype_pb_map()[self._mime_type]
-        except KeyError:
-            raise BadInput(
-                f"{self._mime_type} doesn't have a corresponding File 'kind'",
-            ) from None
+        return pb_v1alpha1.File(kind=kind, content=ret.getvalue())
 
-        return pb.File(kind=kind, content=ret.getvalue())
+    async def to_proto(self, obj: ImageType) -> pb.File:
+        if LazyType["ext.NpNDArray"]("numpy.ndarray").isinstance(obj):
+            image = PIL.Image.fromarray(obj, mode=self._pilmode)
+        elif LazyType["PIL.Image.Image"]("PIL.Image.Image").isinstance(obj):
+            image = obj
+        else:
+            raise BadInput(
+                f"Unsupported Image type received: '{type(obj)}', the Image IO descriptor only supports 'np.ndarray' and 'PIL.Image'.",
+            ) from None
+        ret = io.BytesIO()
+        image.save(ret, format=self._format)
+
+        return pb.File(kind=self._mime_type, content=ret.getvalue())
