@@ -3,7 +3,6 @@ from __future__ import annotations
 import json
 import typing as t
 import logging
-from typing import TYPE_CHECKING
 from functools import lru_cache
 
 from starlette.requests import Request
@@ -15,25 +14,28 @@ from ..utils import LazyLoader
 from ..utils.http import set_cookies
 from ...exceptions import BadInput
 from ...exceptions import InvalidArgument
-from ...exceptions import BentoMLException
 from ...exceptions import UnprocessableEntity
+from ...grpc.utils import import_generated_stubs
+from ...grpc.utils import LATEST_PROTOCOL_VERSION
 from ..service.openapi import SUCCESS_DESCRIPTION
 from ..service.openapi.specification import Schema
 from ..service.openapi.specification import MediaType
 
-if TYPE_CHECKING:
+if t.TYPE_CHECKING:
     import numpy as np
-    from typing_extensions import Self
+    import pyarrow
+    import pyspark.sql.types
+    from google.protobuf import message as _message
 
-    from bentoml.grpc.v1alpha1 import service_pb2 as pb
+    from bentoml.grpc.v1 import service_pb2 as pb
+    from bentoml.grpc.v1alpha1 import service_pb2 as pb_v1alpha1
 
     from .. import external_typing as ext
     from .base import OpenAPIResponse
     from ..context import InferenceApiContext as Context
 else:
-    from bentoml.grpc.utils import import_generated_stubs
-
-    pb, _ = import_generated_stubs()
+    pb, _ = import_generated_stubs("v1")
+    pb_v1alpha1, _ = import_generated_stubs("v1alpha1")
     np = LazyLoader("np", globals(), "numpy")
 
 logger = logging.getLogger(__name__)
@@ -60,8 +62,25 @@ FIELDPB_TO_NPDTYPE_NAME_MAP = {
 }
 
 
-@lru_cache(maxsize=1)
-def dtypepb_to_npdtype_map() -> dict[pb.NDArray.DType.ValueType, ext.NpDTypeLike]:
+@t.overload
+def dtypepb_to_npdtype_map(
+    version: t.Literal["v1"] = ...,
+) -> dict[pb.NDArray.DType.ValueType, ext.NpDTypeLike]:
+    ...
+
+
+@t.overload
+def dtypepb_to_npdtype_map(
+    version: t.Literal["v1alpha1"] = ...,
+) -> dict[pb_v1alpha1.NDArray.DType.ValueType, ext.NpDTypeLike]:
+    ...
+
+
+@lru_cache(maxsize=2)
+def dtypepb_to_npdtype_map(
+    version: str = LATEST_PROTOCOL_VERSION,
+) -> dict[int, ext.NpDTypeLike]:
+    pb, _ = import_generated_stubs(version)
     # pb.NDArray.Dtype -> np.dtype
     return {
         pb.NDArray.DTYPE_FLOAT: np.dtype("float32"),
@@ -75,9 +94,26 @@ def dtypepb_to_npdtype_map() -> dict[pb.NDArray.DType.ValueType, ext.NpDTypeLike
     }
 
 
-@lru_cache(maxsize=1)
-def dtypepb_to_fieldpb_map() -> dict[pb.NDArray.DType.ValueType, str]:
-    return {k: npdtype_to_fieldpb_map()[v] for k, v in dtypepb_to_npdtype_map().items()}
+@t.overload
+def dtypepb_to_fieldpb_map(
+    version: t.Literal["v1"] = ...,
+) -> dict[pb.NDArray.DType.ValueType, str]:
+    ...
+
+
+@t.overload
+def dtypepb_to_fieldpb_map(
+    version: t.Literal["v1alpha1"] = ...,
+) -> dict[pb_v1alpha1.NDArray.DType.ValueType, str]:
+    ...
+
+
+@lru_cache(maxsize=2)
+def dtypepb_to_fieldpb_map(version: str = LATEST_PROTOCOL_VERSION) -> dict[int, str]:
+    return {
+        k: npdtype_to_fieldpb_map()[v]
+        for k, v in dtypepb_to_npdtype_map(version).items()
+    }
 
 
 @lru_cache(maxsize=1)
@@ -86,10 +122,26 @@ def fieldpb_to_npdtype_map() -> dict[str, ext.NpDTypeLike]:
     return {k: np.dtype(v) for k, v in FIELDPB_TO_NPDTYPE_NAME_MAP.items()}
 
 
-@lru_cache(maxsize=1)
-def npdtype_to_dtypepb_map() -> dict[ext.NpDTypeLike, pb.NDArray.DType.ValueType]:
+@t.overload
+def npdtype_to_dtypepb_map(
+    version: t.Literal["v1"] = ...,
+) -> dict[ext.NpDTypeLike, pb.NDArray.DType.ValueType]:
+    ...
+
+
+@t.overload
+def npdtype_to_dtypepb_map(
+    version: t.Literal["v1alpha1"] = ...,
+) -> dict[ext.NpDTypeLike, pb_v1alpha1.NDArray.DType.ValueType]:
+    ...
+
+
+@lru_cache(maxsize=2)
+def npdtype_to_dtypepb_map(
+    version: str = LATEST_PROTOCOL_VERSION,
+) -> dict[ext.NpDTypeLike, int]:
     # np.dtype -> pb.NDArray.Dtype
-    return {v: k for k, v in dtypepb_to_npdtype_map().items()}
+    return {v: k for k, v in dtypepb_to_npdtype_map(version).items()}
 
 
 @lru_cache(maxsize=1)
@@ -116,7 +168,9 @@ def _is_matched_shape(left: tuple[int, ...], right: tuple[int, ...]) -> bool:
 
 # TODO: when updating docs, add examples with gRPCurl
 class NumpyNdarray(
-    IODescriptor["ext.NpNDArray"], descriptor_id="bentoml.io.NumpyNdarray"
+    IODescriptor["ext.NpNDArray"],
+    descriptor_id="bentoml.io.NumpyNdarray",
+    proto_fields=("ndarray",),
 ):
     """
     :obj:`NumpyNdarray` defines API specification for the inputs/outputs of a Service, where
@@ -206,7 +260,6 @@ class NumpyNdarray(
         :obj:`~bentoml._internal.io_descriptors.IODescriptor`: IO Descriptor that represents a :code:`np.ndarray`.
     """
 
-    _proto_fields = ("ndarray",)
     _mime_type = "application/json"
 
     def __init__(
@@ -252,7 +305,7 @@ class NumpyNdarray(
         }
 
     @classmethod
-    def from_spec(cls, spec: dict[str, t.Any]) -> Self:
+    def from_spec(cls, spec: dict[str, t.Any]) -> t.Self:
         if "args" not in spec:
             raise InvalidArgument(f"Missing args key in NumpyNdarray spec: {spec}")
         res = NumpyNdarray(**spec["args"])
@@ -381,10 +434,11 @@ class NumpyNdarray(
 
     def _from_sample(self, sample: ext.NpNDArray | t.Sequence[t.Any]) -> ext.NpNDArray:
         """
-        Create a :obj:`NumpyNdarray` IO Descriptor from given inputs.
+        Create a :class:`~bentoml._internal.io_descriptors.numpy.NumpyNdarray` IO Descriptor from given inputs.
 
         Args:
-            sample: Given sample ``np.ndarray`` data
+            sample: Given sample ``np.ndarray`` data. It also accepts a sequence-like data type that
+                    can be converted to ``np.ndarray``.
             enforce_dtype: Enforce a certain data type. :code:`dtype` must be specified at function
                            signature. If you don't want to enforce a specific dtype then change
                            :code:`enforce_dtype=False`.
@@ -393,7 +447,7 @@ class NumpyNdarray(
                            :code:`enforce_shape=False`.
 
         Returns:
-            :obj:`NumpyNdarray`: :code:`NumpyNdarray` IODescriptor from given users inputs.
+            :class:`~bentoml._internal.io_descriptors.numpy.NumpyNdarray`: IODescriptor from given users inputs.
 
         Example:
 
@@ -417,20 +471,25 @@ class NumpyNdarray(
            @svc.api(input=input_spec, output=NumpyNdarray())
            async def predict(input: NDArray[np.int16]) -> NDArray[Any]:
                return await runner.async_run(input)
+
+        Raises:
+            :class:`BadInput`: If given sample is a type ``numpy.generic``. This exception
+                                       will also be raised if we failed to create a ``np.ndarray``
+                                       from given sample.
         """
         if isinstance(sample, np.generic):
-            raise BentoMLException(
+            raise BadInput(
                 "'NumpyNdarray.from_sample()' expects a 'numpy.array', not 'numpy.generic'."
             ) from None
         try:
             if not isinstance(sample, np.ndarray):
                 sample = np.array(sample)
         except ValueError:
-            raise BentoMLException(
-                f"Failed to create a 'numpy.ndarray' from given sample {sample}"
-            ) from None
-        self._dtype = sample.dtype
-        self._shape = sample.shape
+            raise BadInput(f"Given sample ({sample}) is not a numpy ND-array") from None
+        if self._dtype is None:
+            self._dtype = sample.dtype
+        if self._shape is None:
+            self._shape = sample.shape
         return sample
 
     async def from_proto(self, field: pb.NDArray | bytes) -> ext.NpNDArray:
@@ -455,7 +514,6 @@ class NumpyNdarray(
            Serialized bytes has a specialized bytes representation and should not be used by users directly.
         """
         if isinstance(field, bytes):
-
             # We will be using ``np.frombuffer`` to deserialize the bytes.
             # This means that we need to ensure that ``dtype`` are provided to the IO descriptor
             #
@@ -476,7 +534,14 @@ class NumpyNdarray(
             dtype: ext.NpDTypeLike = self._dtype
             array = np.frombuffer(field, dtype=self._dtype)
         else:
-            assert isinstance(field, pb.NDArray)
+            if isinstance(field, pb_v1alpha1.NDArray):
+                version = "v1alpha1"
+            elif isinstance(field, pb.NDArray):
+                version = "v1"
+            else:
+                raise BadInput(
+                    f"Expected 'pb.NDArray' or 'bytes' but received {type(field)}"
+                ) from None
 
             # The behaviour of dtype are as follows:
             # - if not provided:
@@ -489,11 +554,13 @@ class NumpyNdarray(
                 dtype = None
             else:
                 try:
-                    dtype = dtypepb_to_npdtype_map()[field.dtype]
+                    dtype = dtypepb_to_npdtype_map(version=version)[field.dtype]
                 except KeyError:
                     raise BadInput(f"{field.dtype} is invalid.") from None
             if dtype is not None:
-                values_array = getattr(field, dtypepb_to_fieldpb_map()[field.dtype])
+                values_array = getattr(
+                    field, dtypepb_to_fieldpb_map(version=version)[field.dtype]
+                )
             else:
                 fieldpb = [
                     f.name for f, _ in field.ListFields() if f.name.endswith("_values")
@@ -523,7 +590,21 @@ class NumpyNdarray(
         # We will try to run validation process before sending this of to the user.
         return self.validate_array(array)
 
-    async def to_proto(self, obj: ext.NpNDArray) -> pb.NDArray:
+    @t.overload
+    async def _to_proto_impl(
+        self, obj: ext.NpNDArray, *, version: t.Literal["v1"]
+    ) -> pb.NDArray:
+        ...
+
+    @t.overload
+    async def _to_proto_impl(
+        self, obj: ext.NpNDArray, *, version: t.Literal["v1alpha1"]
+    ) -> pb_v1alpha1.NDArray:
+        ...
+
+    async def _to_proto_impl(
+        self, obj: ext.NpNDArray, *, version: str
+    ) -> _message.Message:
         """
         Process given objects and convert it to grpc protobuf response.
 
@@ -538,9 +619,11 @@ class NumpyNdarray(
         except BadInput as e:
             raise e from None
 
+        pb, _ = import_generated_stubs(version)
+
         try:
             fieldpb = npdtype_to_fieldpb_map()[obj.dtype]
-            dtypepb = npdtype_to_dtypepb_map()[obj.dtype]
+            dtypepb = npdtype_to_dtypepb_map(version=version)[obj.dtype]
             return pb.NDArray(
                 dtype=dtypepb,
                 shape=tuple(obj.shape),
@@ -550,3 +633,46 @@ class NumpyNdarray(
             raise BadInput(
                 f"Unsupported dtype '{obj.dtype}' for response message.",
             ) from None
+
+    def from_arrow(self, batch: pyarrow.RecordBatch) -> ext.NpNDArray:
+        df = batch.to_pandas()
+        if not LazyType("pandas", "DataFrame").isinstance(df):
+            raise InvalidArgument("Unable to convert input into numpy ndarray.")
+        return df.to_numpy()
+
+    def to_arrow(self, arr: ext.NpNDArray) -> pyarrow.RecordBatch:
+        import pyarrow
+
+        return pyarrow.RecordBatch.from_arrays([pyarrow.array(arr)], names=["output"])
+
+    def spark_schema(self) -> pyspark.sql.types.StructType:
+        from pyspark.sql.types import StructType
+        from pyspark.sql.types import StructField
+        from pyspark.pandas.typedef import as_spark_type
+
+        if self._dtype is None:
+            raise InvalidArgument(
+                "Cannot perform batch inference with a numpy output without a known dtype; please provide a dtype."
+            )
+        if self._shape is None:
+            raise InvalidArgument(
+                "Cannot perform batch inference with a numpy output without a known shape; please provide a shape."
+            )
+        if len(self._shape) != 1:
+            raise InvalidArgument(
+                "Cannot perform batch inference with a multidimensional numpy ndarray output; consider using pandas DataFrames instead."
+            )
+
+        try:
+            out_spark_type = as_spark_type(self._dtype)
+        except TypeError:
+            raise InvalidArgument(
+                f"dtype {self._dtype} is not supported for batch inference."
+            )
+        return StructType([StructField("out", out_spark_type, nullable=False)])
+
+    async def to_proto(self, obj: ext.NpNDArray) -> pb.NDArray:
+        return await self._to_proto_impl(obj, version="v1")
+
+    async def to_proto_v1alpha1(self, obj: ext.NpNDArray) -> pb_v1alpha1.NDArray:
+        return await self._to_proto_impl(obj, version="v1alpha1")
