@@ -6,7 +6,6 @@ import typing as t
 import logging
 import functools
 from enum import Enum
-from typing import TYPE_CHECKING
 from concurrent.futures import ThreadPoolExecutor
 
 from starlette.requests import Request
@@ -20,6 +19,7 @@ from ...exceptions import BadInput
 from ...exceptions import InvalidArgument
 from ...exceptions import UnprocessableEntity
 from ...exceptions import MissingDependencyException
+from ...grpc.utils import import_generated_stubs
 from ..service.openapi import SUCCESS_DESCRIPTION
 from ..utils.lazy_loader import LazyLoader
 from ..service.openapi.specification import Schema
@@ -27,21 +27,23 @@ from ..service.openapi.specification import MediaType
 
 EXC_MSG = "pandas' is required to use PandasDataFrame or PandasSeries. Install with 'pip install bentoml[io-pandas]'"
 
-if TYPE_CHECKING:
+if t.TYPE_CHECKING:
     import numpy as np
     import pandas as pd
-    from typing_extensions import Self
+    import pyarrow
+    import pyspark.sql.types
+    from google.protobuf import message as _message
 
-    from bentoml.grpc.v1alpha1 import service_pb2 as pb
+    from bentoml.grpc.v1 import service_pb2 as pb
+    from bentoml.grpc.v1alpha1 import service_pb2 as pb_v1alpha1
 
     from .. import external_typing as ext
     from .base import OpenAPIResponse
     from ..context import InferenceApiContext as Context
 
 else:
-    from bentoml.grpc.utils import import_generated_stubs
-
-    pb, _ = import_generated_stubs()
+    pb, _ = import_generated_stubs("v1")
+    pb_v1alpha1, _ = import_generated_stubs("v1alpha1")
     pd = LazyLoader("pd", globals(), "pandas", exc_msg=EXC_MSG)
     np = LazyLoader("np", globals(), "numpy")
 
@@ -197,7 +199,9 @@ def _validate_serialization_format(serialization_format: SerializationFormat):
 
 
 class PandasDataFrame(
-    IODescriptor["ext.PdDataFrame"], descriptor_id="bentoml.io.PandasDataFrame"
+    IODescriptor["ext.PdDataFrame"],
+    descriptor_id="bentoml.io.PandasDataFrame",
+    proto_fields=("dataframe",),
 ):
     """
     :obj:`PandasDataFrame` defines API specification for the inputs/outputs of a Service,
@@ -312,8 +316,6 @@ class PandasDataFrame(
         :obj:`PandasDataFrame`: IO Descriptor that represents a :code:`pd.DataFrame`.
     """
 
-    _proto_fields = ("dataframe",)
-
     def __init__(
         self,
         orient: ext.DataFrameOrient = "records",
@@ -340,7 +342,7 @@ class PandasDataFrame(
 
     def _from_sample(self, sample: ext.PdDataFrame) -> ext.PdDataFrame:
         """
-        Create a :obj:`PandasDataFrame` IO Descriptor from given inputs.
+        Create a :class:`~bentoml._internal.io_descriptors.pandas.PandasDataFrame` IO Descriptor from given inputs.
 
         Args:
             sample: Given sample ``pd.DataFrame`` data
@@ -371,7 +373,7 @@ class PandasDataFrame(
                             - :obj:`csv` - CSV text format (inferred from content-type ``"text/csv"``)
 
         Returns:
-            :obj:`PandasDataFrame`: :code:`PandasDataFrame` IODescriptor from given users inputs.
+            :class:`~bentoml._internal.io_descriptors.pandas.PandasDataFrame`: IODescriptor from given users inputs.
 
         Example:
 
@@ -387,8 +389,16 @@ class PandasDataFrame(
            def predict(inputs: pd.DataFrame) -> pd.DataFrame: ...
         """
         if LazyType["ext.NpNDArray"]("numpy", "ndarray").isinstance(sample):
+            logger.warning(
+                "'from_sample' from type '%s' is deprecated. Make sure to only pass pandas DataFrame.",
+                type(sample),
+            )
             sample = pd.DataFrame(sample)
         elif isinstance(sample, str):
+            logger.warning(
+                "'from_sample' from type '%s' is deprecated. Make sure to only pass pandas DataFrame.",
+                type(sample),
+            )
             try:
                 if os.path.exists(sample):
                     try:
@@ -416,10 +426,12 @@ class PandasDataFrame(
                 raise InvalidArgument(
                     f"Failed to create a 'pd.DataFrame' from sample {sample}: {e}"
                 ) from None
-        self._shape = sample.shape
-        self._columns = [str(i) for i in list(sample.columns)]
+        if self._shape is None:
+            self._shape = sample.shape
+        if self._columns is None:
+            self._columns = [str(i) for i in list(sample.columns)]
         if self._dtype is None:
-            self._dtype = True  # infer dtype automatically
+            self._dtype = sample.dtypes
         return sample
 
     def _convert_dtype(
@@ -453,7 +465,7 @@ class PandasDataFrame(
         }
 
     @classmethod
-    def from_spec(cls, spec: dict[str, t.Any]) -> Self:
+    def from_spec(cls, spec: dict[str, t.Any]) -> t.Self:
         if "args" not in spec:
             raise InvalidArgument(f"Missing args key in PandasDataFrame spec: {spec}")
         res = PandasDataFrame(**spec["args"])
@@ -516,14 +528,15 @@ class PandasDataFrame(
         _validate_serialization_format(serialization_format)
 
         obj = await request.body()
+        dtype = self._dtype if self._enforce_dtype else None
         if serialization_format is SerializationFormat.JSON:
-            assert not isinstance(self._dtype, bool)
-            res = pd.read_json(io.BytesIO(obj), dtype=self._dtype, orient=self._orient)
+            if dtype is None:
+                dtype = True  # infer dtype automatically
+            res = pd.read_json(io.BytesIO(obj), dtype=dtype, orient=self._orient)
         elif serialization_format is SerializationFormat.PARQUET:
             res = pd.read_parquet(io.BytesIO(obj), engine=get_parquet_engine())
         elif serialization_format is SerializationFormat.CSV:
-            assert not isinstance(self._dtype, bool)
-            res: ext.PdDataFrame = pd.read_csv(io.BytesIO(obj), dtype=self._dtype)
+            res: ext.PdDataFrame = pd.read_csv(io.BytesIO(obj), dtype=dtype)
         else:
             raise InvalidArgument(
                 f"Unknown serialization format ({serialization_format})."
@@ -581,7 +594,6 @@ class PandasDataFrame(
     def validate_dataframe(
         self, dataframe: ext.PdDataFrame, exception_cls: t.Type[Exception] = BadInput
     ) -> ext.PdDataFrame:
-
         if not LazyType["ext.PdDataFrame"]("pandas.core.frame.DataFrame").isinstance(
             dataframe
         ):
@@ -595,13 +607,18 @@ class PandasDataFrame(
         #     if self._enforce_dtype:
         #         raise exception_cls(msg) from None
 
-        if self._columns is not None and len(self._columns) != dataframe.shape[1]:
-            msg = "length of 'columns' (%d) does not match the # of columns of incoming data."
-            if self._apply_column_names:
-                raise BadInput(msg % len(self._columns)) from None
-            else:
-                logger.debug(msg, len(self._columns))
-                dataframe.columns = pd.Index(self._columns)
+        if self._apply_column_names:
+            if not self._columns:
+                raise BadInput(
+                    "When apply_column_names is set, you must provide columns."
+                )
+
+            if len(self._columns) != dataframe.shape[1]:
+                raise BadInput(
+                    f"length of 'columns' ({len(self._columns)}) does not match the # of columns of incoming data ({dataframe.shape[1]})."
+                ) from None
+
+            dataframe.columns = pd.Index(self._columns)
 
         # TODO: convert from wide to long format (melt())
         if self._shape is not None and self._shape != dataframe.shape:
@@ -638,7 +655,6 @@ class PandasDataFrame(
             # dtype of given fields per Series to match with types of a given
             # columns, hence, this would result in a wrong DataFrame that is not
             # expected by our users.
-            assert isinstance(field, pb.DataFrame)
             # columns orient: { column_name : {index : columns.series._value}}
             if self._orient != "columns":
                 raise BadInput(
@@ -664,7 +680,21 @@ class PandasDataFrame(
             )
         return self.validate_dataframe(dataframe)
 
-    async def to_proto(self, obj: ext.PdDataFrame) -> pb.DataFrame:
+    @t.overload
+    async def _to_proto_impl(
+        self, obj: ext.PdDataFrame, *, version: t.Literal["v1"]
+    ) -> pb.DataFrame:
+        ...
+
+    @t.overload
+    async def _to_proto_impl(
+        self, obj: ext.PdDataFrame, *, version: t.Literal["v1alpha1"]
+    ) -> pb_v1alpha1.DataFrame:
+        ...
+
+    async def _to_proto_impl(
+        self, obj: ext.PdDataFrame, *, version: str
+    ) -> _message.Message:
         """
         Process given objects and convert it to grpc protobuf response.
 
@@ -677,6 +707,8 @@ class PandasDataFrame(
         """
         from .numpy import npdtype_to_fieldpb_map
 
+        pb, _ = import_generated_stubs(version)
+
         # TODO: support different serialization format
         obj = self.validate_dataframe(obj)
         mapping = npdtype_to_fieldpb_map()
@@ -704,12 +736,57 @@ class PandasDataFrame(
             ],
         )
 
+    def from_arrow(self, batch: pyarrow.RecordBatch) -> ext.PdDataFrame:
+        res = batch.to_pandas()
+        if isinstance(res, pd.DataFrame):
+            return res
+        if isinstance(res, pd.Series):
+            return res.to_frame()
+
+    def to_arrow(self, df: pd.Series[t.Any]) -> pyarrow.RecordBatch:
+        import pyarrow
+
+        return pyarrow.RecordBatch.from_pandas(df)
+
+    def spark_schema(self) -> pyspark.sql.types.StructType:
+        from pyspark.sql.types import StructType
+        from pyspark.sql.types import StructField
+        from pyspark.pandas.typedef import as_spark_type
+
+        if self._dtype is None or self._dtype:
+            raise InvalidArgument(
+                "Cannot perform batch inference with a numpy output without a known dtype; please provide a dtype."
+            )
+
+        if isinstance(self._dtype, dict):
+            fields = []
+            for col_name, col_type in self._dtype:
+                try:
+                    fields.append(StructField(col_name, as_spark_type(col_type)))
+                except TypeError:
+                    raise InvalidArgument(
+                        f"dtype {col_type} is not supported for batch inference."
+                    )
+            return StructType(fields)
+        else:
+            raise NotImplementedError(
+                "Only dict dtypes are currently supported for dataframes"
+            )
+
+    async def to_proto(self, obj: ext.PdDataFrame) -> pb.DataFrame:
+        return await self._to_proto_impl(obj, version="v1")
+
+    async def to_proto_v1alpha1(self, obj: ext.PdDataFrame) -> pb_v1alpha1.DataFrame:
+        return await self._to_proto_impl(obj, version="v1alpha1")
+
 
 class PandasSeries(
-    IODescriptor["ext.PdSeries"], descriptor_id="bentoml.io.PandasSeries"
+    IODescriptor["ext.PdSeries"],
+    descriptor_id="bentoml.io.PandasSeries",
+    proto_fields=("series",),
 ):
     """
-    :code:`PandasSeries` defines API specification for the inputs/outputs of a Service, where
+    ``PandasSeries`` defines API specification for the inputs/outputs of a Service, where
     either inputs will be converted to or outputs will be converted from type
     :code:`pd.Series` as specified in your API function signature.
 
@@ -775,7 +852,6 @@ class PandasSeries(
                 - :obj:`columns` - :code:`dict[str, Any]` ↦ {``column`` ↠ {``index`` ↠ ``value``}}
                 - :obj:`values` - :code:`dict[str, Any]` ↦ Values arrays
         columns: List of columns name that users wish to update.
-        apply_column_names (`bool`, `optional`, default to :code:`False`):
         apply_column_names: Whether to update incoming DataFrame columns. If :code:`apply_column_names=True`,
                             then ``columns`` must be specified.
         dtype: Data type users wish to convert their inputs/outputs to. If it is a boolean,
@@ -786,15 +862,16 @@ class PandasSeries(
         shape: Optional shape check that users can specify for their incoming HTTP
                requests. We will only check the number of columns you specified for your
                given shape:
+
                .. code-block:: python
                   :caption: `service.py`
 
-                  import pandas as pd
-                  from bentoml.io import PandasSeries
+                   import pandas as pd
+                   from bentoml.io import PandasSeries
 
-                  @svc.api(input=PandasSeries(shape=(51,), enforce_shape=True), output=PandasSeries())
-                  def infer(input_series: pd.Series) -> pd.Series:
-                  # if input_series has shape (40,), it will error
+                   @svc.api(input=PandasSeries(shape=(51,), enforce_shape=True), output=PandasSeries())
+                   def infer(input_series: pd.Series) -> pd.Series:
+                        # if input_series has shape (40,), it will error
                         ...
         enforce_shape: Whether to enforce a certain shape. If ``enforce_shape=True`` then ``shape`` must be specified.
 
@@ -802,7 +879,6 @@ class PandasSeries(
         :obj:`PandasSeries`: IO Descriptor that represents a :code:`pd.Series`.
     """
 
-    _proto_fields = ("series",)
     _mime_type = "application/json"
 
     def __init__(
@@ -821,7 +897,7 @@ class PandasSeries(
 
     def _from_sample(self, sample: ext.PdSeries | t.Sequence[t.Any]) -> ext.PdSeries:
         """
-        Create a :obj:`PandasSeries` IO Descriptor from given inputs.
+        Create a :class:`~bentoml._internal.io_descriptors.pandas.PandasSeries` IO Descriptor from given inputs.
 
         Args:
             sample_input: Given sample ``pd.DataFrame`` data
@@ -841,7 +917,7 @@ class PandasSeries(
                            ``enforce_shape=False``.
 
         Returns:
-            :obj:`PandasSeries`: :code:`PandasSeries` IODescriptor from given users inputs.
+            :class:`~bentoml._internal.io_descriptors.pandas.PandasSeries`: IODescriptor from given users inputs.
 
         Example:
 
@@ -859,8 +935,10 @@ class PandasSeries(
         """
         if not isinstance(sample, pd.Series):
             sample = pd.Series(sample)
-        self._dtype = sample.dtype
-        self._shape = sample.shape
+        if self._dtype is None:
+            self._dtype = sample.dtype
+        if self._shape is None:
+            self._shape = sample.shape
         return sample
 
     def input_type(self) -> LazyType[ext.PdSeries]:
@@ -895,7 +973,7 @@ class PandasSeries(
         }
 
     @classmethod
-    def from_spec(cls, spec: dict[str, t.Any]) -> Self:
+    def from_spec(cls, spec: dict[str, t.Any]) -> t.Self:
         if "args" not in spec:
             raise InvalidArgument(f"Missing args key in PandasSeries spec: {spec}")
         res = PandasSeries(**spec["args"])
@@ -947,7 +1025,7 @@ class PandasSeries(
             io.BytesIO(obj),
             typ="series",
             orient=self._orient,
-            dtype=self._dtype,
+            dtype=self._dtype if self._enforce_dtype else True,
         )
         return self.validate_series(res)
 
@@ -1028,7 +1106,6 @@ class PandasSeries(
                 'Currently not yet implemented. Use "series" instead.'
             )
         else:
-            assert isinstance(field, pb.Series)
             # The behaviour of `from_proto` will mimic the behaviour of `NumpyNdArray.from_proto`,
             # where we will respect self._dtype if set.
             # since self._dtype uses numpy dtype, we will use some of numpy logics here.
@@ -1060,7 +1137,21 @@ class PandasSeries(
 
         return self.validate_series(series)
 
-    async def to_proto(self, obj: ext.PdSeries) -> pb.Series:
+    @t.overload
+    async def _to_proto_impl(
+        self, obj: ext.PdSeries, *, version: t.Literal["v1"]
+    ) -> pb.Series:
+        ...
+
+    @t.overload
+    async def _to_proto_impl(
+        self, obj: ext.PdSeries, *, version: t.Literal["v1alpha1"]
+    ) -> pb_v1alpha1.Series:
+        ...
+
+    async def _to_proto_impl(
+        self, obj: ext.PdSeries, *, version: str
+    ) -> _message.Message:
         """
         Process given objects and convert it to grpc protobuf response.
 
@@ -1073,6 +1164,8 @@ class PandasSeries(
         """
         from .numpy import npdtype_to_fieldpb_map
 
+        pb, _ = import_generated_stubs(version)
+
         try:
             obj = self.validate_series(obj, exception_cls=InvalidArgument)
         except InvalidArgument as e:
@@ -1083,7 +1176,7 @@ class PandasSeries(
         # User shouldn't use mixed dtype in the first place.
         if obj.dtype.kind == "O":
             raise InvalidArgument(
-                f"Series has mixed dtype. Please convert it to a single dtype."
+                "Series has mixed dtype. Please convert it to a single dtype."
             ) from None
         try:
             fieldpb = npdtype_to_fieldpb_map()[obj.dtype]
@@ -1092,3 +1185,46 @@ class PandasSeries(
             raise InvalidArgument(
                 f"Unsupported dtype '{obj.dtype}' for response message."
             ) from None
+
+    def from_arrow(self, batch: pyarrow.RecordBatch) -> pd.Series[t.Any]:
+        res = batch.to_pandas()
+        if isinstance(res, pd.Series):
+            return res
+        if isinstance(res, pd.DataFrame):
+            if len(res.columns) == 1:
+                return res.squeeze()
+            else:
+                raise InvalidArgument(
+                    "Multi-column dataframe was passed when trying to convert to a series."
+                )
+
+    def to_arrow(self, series: pd.Series[t.Any]) -> pyarrow.RecordBatch:
+        import pyarrow
+
+        df = series.to_frame()
+        return pyarrow.RecordBatch.from_pandas(df)
+
+    def spark_schema(self) -> pyspark.sql.types.StructType:
+        from pyspark.sql.types import StructType
+        from pyspark.sql.types import StructField
+        from pyspark.pandas.typedef import as_spark_type
+
+        if self._dtype is None or self._dtype is True:
+            raise InvalidArgument(
+                "Cannot perform batch inference with a pandas series output without a known dtype; please provide a dtype."
+            )
+
+        try:
+            out_spark_type = as_spark_type(self._dtype)
+        except TypeError:
+            raise InvalidArgument(
+                f"dtype {self._dtype} is not supported for batch inference."
+            )
+
+        return StructType([StructField("out", out_spark_type)])
+
+    async def to_proto(self, obj: ext.PdSeries) -> pb.Series:
+        return await self._to_proto_impl(obj, version="v1")
+
+    async def to_proto_v1alpha1(self, obj: ext.PdSeries) -> pb_v1alpha1.Series:
+        return await self._to_proto_impl(obj, version="v1alpha1")
