@@ -18,9 +18,9 @@ import fs.errors
 import fs.mirror
 import cloudpickle  # type: ignore (no cloudpickle types)
 from fs.base import FS
-from cattr.gen import override  # type: ignore (incomplete cattr types)
-from cattr.gen import make_dict_structure_fn  # type: ignore (incomplete cattr types)
-from cattr.gen import make_dict_unstructure_fn  # type: ignore (incomplete cattr types)
+from cattr.gen import override
+from cattr.gen import make_dict_structure_fn
+from cattr.gen import make_dict_unstructure_fn
 from simple_di import inject
 from simple_di import Provide
 
@@ -31,6 +31,7 @@ from ..types import MetadataDict
 from ..utils import bentoml_cattr
 from ..utils import label_validator
 from ..utils import metadata_validator
+from ..utils import normalize_labels_value
 from ...exceptions import NotFound
 from ...exceptions import BentoMLException
 from ..configuration import BENTOML_VERSION
@@ -70,6 +71,11 @@ class ModelOptions:
         return attr.asdict(self)
 
 
+@attr.define
+class PartialKwargsModelOptions(ModelOptions):
+    partial_kwargs: t.Dict[str, t.Any] = attr.field(factory=dict)
+
+
 @attr.define(repr=False, eq=False, init=False)
 class Model(StoreItem):
     _tag: Tag
@@ -200,14 +206,9 @@ class Model(StoreItem):
 
     @inject
     def save(
-        self, model_store: ModelStore = Provide[BentoMLContainer.model_store]
+        self,
+        model_store: ModelStore = Provide[BentoMLContainer.model_store],
     ) -> Model:
-        self._save(model_store)
-
-        logger.info("Successfully saved %s", self)
-        return self
-
-    def _save(self, model_store: ModelStore) -> Model:
         try:
             self.validate()
         except BentoMLException as e:
@@ -235,7 +236,7 @@ class Model(StoreItem):
         try:
             res.validate()
         except BentoMLException as e:
-            raise BentoMLException(f"Failed to load {self!s}: {e}") from None
+            raise BentoMLException(f"Failed to load {res!s}: {e}") from None
 
         return res
 
@@ -496,7 +497,6 @@ bentoml_cattr.register_unstructure_hook(
 
 @attr.define(repr=False, eq=False, frozen=True)
 class ModelInfo:
-
     # for backward compatibility in case new fields are added to BentoInfo.
     __forbid_extra_keys__ = False
     # omit field in yaml file if it is not provided by the user.
@@ -506,7 +506,9 @@ class ModelInfo:
     name: str
     version: str
     module: str
-    labels: t.Dict[str, str] = attr.field(validator=label_validator)
+    labels: t.Dict[str, str] = attr.field(
+        validator=label_validator, converter=normalize_labels_value
+    )
     _options: t.Dict[str, t.Any]
     metadata: MetadataDict = attr.field(validator=metadata_validator, converter=dict)
     context: ModelContext = attr.field()
@@ -614,7 +616,7 @@ class ModelInfo:
         return self._cached_options
 
     def to_dict(self) -> t.Dict[str, t.Any]:
-        return bentoml_cattr.unstructure(self)  # type: ignore (incomplete cattr types)
+        return bentoml_cattr.unstructure(self)
 
     @overload
     def dump(self, stream: io.StringIO) -> io.BytesIO:
@@ -675,9 +677,9 @@ bentoml_cattr.register_structure_hook_func(
     ),
 )
 bentoml_cattr.register_unstructure_hook_func(
-    lambda cls: issubclass(cls, ModelInfo),  # type: ignore (lambda)
+    lambda cls: issubclass(cls, ModelInfo),
     # Ignore tag, tag is saved via the name and version field
-    make_dict_unstructure_fn(  # type: ignore (incomplete cattr types)
+    make_dict_unstructure_fn(
         ModelInfo,
         bentoml_cattr,
         tag=override(omit=True),
