@@ -1,22 +1,42 @@
 from __future__ import annotations
 
 import abc
-import pickle
+import sys
+import base64
 import typing as t
 import itertools
-from typing import TYPE_CHECKING
 
 from simple_di import inject
 from simple_di import Provide
 
 from ..types import LazyType
+from ..utils import LazyLoader
 from ..configuration.containers import BentoMLContainer
 
+if sys.version_info < (3, 8):
+    import pickle5 as pickle
+else:
+    import pickle
+
+
 SingleType = t.TypeVar("SingleType")
 BatchType = t.TypeVar("BatchType")
 
-if TYPE_CHECKING:
+TRITON_EXC_MSG = "tritonclient is required to use triton with BentoML. Install with 'pip install \"tritonclient[all]>=2.29.0\"'."
+
+if t.TYPE_CHECKING:
+    import tritonclient.grpc.aio as tritongrpcclient
+    import tritonclient.http.aio as tritonhttpclient
+
     from .. import external_typing as ext
+else:
+    np = LazyLoader("np", globals(), "numpy")
+    tritongrpcclient = LazyLoader(
+        "tritongrpcclient", globals(), "tritonclient.grpc.aio", exc_msg=TRITON_EXC_MSG
+    )
+    tritonhttpclient = LazyLoader(
+        "tritonhttpclient", globals(), "tritonclient.http.aio", exc_msg=TRITON_EXC_MSG
+    )
 
 
 class Payload(t.NamedTuple):
@@ -46,6 +66,71 @@ class DataContainer(t.Generic[SingleType, BatchType]):
     def from_payload(cls, payload: Payload) -> BatchType:
         ...
 
+    @t.overload
+    @classmethod
+    def to_triton_payload(
+        cls,
+        inp: tritongrpcclient.InferInput,
+        meta: tritongrpcclient.service_pb2.ModelMetadataResponse.TensorMetadata,
+        _use_http_client: t.Literal[False] = ...,
+    ) -> tritongrpcclient.InferInput:
+        ...
+
+    @t.overload
+    @classmethod
+    def to_triton_payload(
+        cls,
+        inp: tritonhttpclient.InferInput,
+        meta: dict[str, t.Any],
+        _use_http_client: t.Literal[True] = ...,
+    ) -> tritongrpcclient.InferInput:
+        ...
+
+    @classmethod
+    def to_triton_payload(
+        cls, inp: SingleType, meta: t.Any, _use_http_client: bool = False
+    ) -> t.Any:
+        """
+        Convert given input types to a Triton payload.
+
+        Implementation of this method should always return a InferInput that has
+        data of a Numpy array.
+        """
+        if _use_http_client:
+            return cls.to_triton_http_payload(inp, meta)
+        else:
+            return cls.to_triton_grpc_payload(inp, meta)
+
+    @classmethod
+    def to_triton_grpc_payload(
+        cls,
+        inp: SingleType,
+        meta: tritongrpcclient.service_pb2.ModelMetadataResponse.TensorMetadata,
+    ) -> tritongrpcclient.InferInput:
+        """
+        Convert given input types to a Triton payload via gRPC client.
+
+        Implementation of this method should always return a InferInput.
+        """
+        raise NotImplementedError(
+            f"{cls.__name__} doesn't support converting to Triton payload."
+        )
+
+    @classmethod
+    def to_triton_http_payload(
+        cls,
+        inp: SingleType,
+        meta: dict[str, t.Any],
+    ) -> tritonhttpclient.InferInput:
+        """
+        Convert given input types to a Triton payload via HTTP client.
+
+        Implementation of this method should always return a InferInput.
+        """
+        raise NotImplementedError(
+            f"{cls.__name__} doesn't support converting to Triton payload."
+        )
+
     @classmethod
     @abc.abstractmethod
     def batches_to_batch(
@@ -77,25 +162,84 @@ class DataContainer(t.Generic[SingleType, BatchType]):
         ...
 
 
-class NdarrayContainer(
+class TritonInferInputDataContainer(
     DataContainer[
-        "ext.NpNDArray",
-        "ext.NpNDArray",
+        t.Union["tritongrpcclient.InferInput", "tritonhttpclient.InferInput"],
+        t.Union["tritongrpcclient.InferInput", "tritonhttpclient.InferInput"],
     ]
 ):
+    @classmethod
+    def to_payload(cls, batch: tritongrpcclient.InferInput, batch_dim: int) -> Payload:
+        raise NotImplementedError
+
+    @classmethod
+    def from_payload(cls, payload: Payload) -> tritongrpcclient.InferInput:
+        raise NotImplementedError
+
+    @classmethod
+    def to_triton_grpc_payload(
+        cls,
+        inp: tritongrpcclient.InferInput | tritonhttpclient.InferInput,
+        meta: tritongrpcclient.service_pb2.ModelMetadataResponse.TensorMetadata,
+    ) -> tritongrpcclient.InferInput:
+        return t.cast("tritongrpcclient.InferInput", inp)
+
+    @classmethod
+    def to_triton_http_payload(
+        cls,
+        inp: tritongrpcclient.InferInput | tritonhttpclient.InferInput,
+        meta: dict[str, t.Any],
+    ) -> tritonhttpclient.InferInput:
+        return t.cast("tritonhttpclient.InferInput", inp)
+
+    @classmethod
+    @abc.abstractmethod
+    def batches_to_batch(
+        cls,
+        batches: t.Sequence[tritongrpcclient.InferInput | tritonhttpclient.InferInput],
+        batch_dim: int,
+    ) -> tuple[tritongrpcclient.InferInput | tritonhttpclient.InferInput, list[int]]:
+        raise NotImplementedError
+
+    @classmethod
+    @abc.abstractmethod
+    def batch_to_batches(
+        cls,
+        batch: tritongrpcclient.InferInput | tritonhttpclient.InferInput,
+        indices: t.Sequence[int],
+        batch_dim: int,
+    ) -> list[tritongrpcclient.InferInput | tritonhttpclient.InferInput]:
+        raise NotImplementedError
+
+    @classmethod
+    @abc.abstractmethod
+    def batch_to_payloads(
+        cls,
+        batch: tritongrpcclient.InferInput | tritonhttpclient.InferInput,
+        indices: t.Sequence[int],
+        batch_dim: int,
+    ) -> list[Payload]:
+        raise NotImplementedError
+
+    @classmethod
+    @abc.abstractmethod
+    def from_batch_payloads(
+        cls,
+        payloads: t.Sequence[Payload],
+        batch_dim: int,
+    ) -> tuple[tritongrpcclient.InferInput | tritonhttpclient.InferInput, list[int]]:
+        raise NotImplementedError
+
+
+class NdarrayContainer(DataContainer["ext.NpNDArray", "ext.NpNDArray"]):
     @classmethod
     def batches_to_batch(
         cls,
         batches: t.Sequence[ext.NpNDArray],
         batch_dim: int = 0,
     ) -> tuple[ext.NpNDArray, list[int]]:
-        import numpy as np
-
         # numpy.concatenate may consume lots of memory, need optimization later
-        batch: ext.NpNDArray = np.concatenate(  # type: ignore  (no numpy types)
-            batches,
-            axis=batch_dim,
-        )
+        batch: ext.NpNDArray = np.concatenate(batches, axis=batch_dim)
         indices = list(
             itertools.accumulate(subbatch.shape[batch_dim] for subbatch in batches)
         )
@@ -109,61 +253,90 @@ class NdarrayContainer(
         indices: t.Sequence[int],
         batch_dim: int = 0,
     ) -> list[ext.NpNDArray]:
-        import numpy as np
+        return np.split(batch, indices[1:-1], axis=batch_dim)
 
-        return np.split(batch, indices[1:-1], axis=batch_dim)  # type: ignore  (no numpy types)
+    @classmethod
+    def to_triton_grpc_payload(
+        cls,
+        inp: ext.NpNDArray,
+        meta: tritongrpcclient.service_pb2.ModelMetadataResponse.TensorMetadata,
+    ) -> tritongrpcclient.InferInput:
+        InferInput = tritongrpcclient.InferInput(
+            meta.name, inp.shape, tritongrpcclient.np_to_triton_dtype(inp.dtype)
+        )
+        InferInput.set_data_from_numpy(inp)
+        return InferInput
+
+    @classmethod
+    def to_triton_http_payload(
+        cls,
+        inp: ext.NpNDArray,
+        meta: dict[str, t.Any],
+    ) -> tritonhttpclient.InferInput:
+        InferInput = tritonhttpclient.InferInput(
+            meta["name"], inp.shape, tritonhttpclient.np_to_triton_dtype(inp.dtype)
+        )
+        InferInput.set_data_from_numpy(inp)
+        return InferInput
 
     @classmethod
-    @inject
     def to_payload(
         cls,
         batch: ext.NpNDArray,
         batch_dim: int,
-        plasma_db: ext.PlasmaClient | None = Provide[BentoMLContainer.plasma_db],
     ) -> Payload:
-        if plasma_db:
+
+        # skip 0-dimensional array
+        if batch.shape:
+
+            buffers: list[pickle.PickleBuffer] = []
+            if not (batch.flags["C_CONTIGUOUS"] or batch.flags["F_CONTIGUOUS"]):
+                # TODO: use fortan contiguous if it's faster
+                batch = np.ascontiguousarray(batch)
+            bs = pickle.dumps(batch, protocol=5, buffer_callback=buffers.append)
+            bs_str = base64.b64encode(bs).decode("ascii")
+            buffer_bs = buffers[0].raw().tobytes()
+            # release memory
+            buffers[0].release()
             return cls.create_payload(
-                plasma_db.put(batch).binary(),
+                buffer_bs,
                 batch.shape[batch_dim],
-                {"plasma": True},
+                {
+                    "format": "pickle5",
+                    "pickle_bytes": bs_str,
+                },
             )
 
         return cls.create_payload(
             pickle.dumps(batch),
             batch.shape[batch_dim],
-            {"plasma": False},
+            {"format": "default"},
         )
 
     @classmethod
-    @inject
     def from_payload(
         cls,
         payload: Payload,
-        plasma_db: ext.PlasmaClient | None = Provide[BentoMLContainer.plasma_db],
     ) -> ext.NpNDArray:
-        if payload.meta.get("plasma"):
-            import pyarrow.plasma as plasma
-
-            assert plasma_db
-            return plasma_db.get(plasma.ObjectID(payload.data))
+        format = payload.meta.get("format", "default")
+        if format == "pickle5":
+            bs_str = payload.meta["pickle_bytes"]
+            bs = base64.b64decode(bs_str)
+            recovered_buffers = [pickle.PickleBuffer(payload.data)]
+            return pickle.loads(bs, buffers=recovered_buffers)
 
         return pickle.loads(payload.data)
 
     @classmethod
-    @inject
     def batch_to_payloads(
         cls,
         batch: ext.NpNDArray,
         indices: t.Sequence[int],
         batch_dim: int = 0,
-        plasma_db: ext.PlasmaClient | None = Provide[BentoMLContainer.plasma_db],
     ) -> list[Payload]:
-
         batches = cls.batch_to_batches(batch, indices, batch_dim)
 
-        payloads = [
-            cls.to_payload(subbatch, batch_dim, plasma_db) for subbatch in batches
-        ]
+        payloads = [cls.to_payload(subbatch, batch_dim) for subbatch in batches]
         return payloads
 
     @classmethod
@@ -172,9 +345,8 @@ class NdarrayContainer(
         cls,
         payloads: t.Sequence[Payload],
         batch_dim: int = 0,
-        plasma_db: "ext.PlasmaClient" | None = Provide[BentoMLContainer.plasma_db],
     ) -> t.Tuple["ext.NpNDArray", list[int]]:
-        batches = [cls.from_payload(payload, plasma_db) for payload in payloads]
+        batches = [cls.from_payload(payload) for payload in payloads]
         return cls.batches_to_batch(batches, batch_dim)
 
 
@@ -205,7 +377,6 @@ class PandasDataFrameContainer(
         indices: t.Sequence[int],
         batch_dim: int = 0,
     ) -> list[ext.PdDataFrame]:
-
         assert (
             batch_dim == 0
         ), "PandasDataFrameContainer does not support batch_dim other than 0"
@@ -269,7 +440,6 @@ class PandasDataFrameContainer(
         batch_dim: int = 0,
         plasma_db: ext.PlasmaClient | None = Provide[BentoMLContainer.plasma_db],
     ) -> list[Payload]:
-
         batches = cls.batch_to_batches(batch, indices, batch_dim)
 
         payloads = [
@@ -337,7 +507,6 @@ class DefaultContainer(DataContainer[t.Any, t.List[t.Any]]):
         indices: t.Sequence[int],
         batch_dim: int = 0,
     ) -> list[Payload]:
-
         batches = cls.batch_to_batches(batch, indices, batch_dim)
 
         payloads = [cls.to_payload(subbatch, batch_dim) for subbatch in batches]
@@ -409,24 +578,38 @@ def register_builtin_containers():
     DataContainerRegistry.register_container(
         LazyType("numpy", "ndarray"), LazyType("numpy", "ndarray"), NdarrayContainer
     )
-    # DataContainerRegistry.register_container(np.ndarray, np.ndarray, NdarrayContainer)
-
-    # DataContainerRegistry.register_container(
-    #     LazyType("xgboost", "DMatrix"), LazyType("xgboost", "DMatrix"), DMatrixContainer
-    # )
-
     DataContainerRegistry.register_container(
         LazyType("pandas.core.series", "Series"),
         LazyType("pandas.core.frame", "DataFrame"),
         PandasDataFrameContainer,
     )
-
     DataContainerRegistry.register_container(
         LazyType("pandas.core.frame", "DataFrame"),
         LazyType("pandas.core.frame", "DataFrame"),
         PandasDataFrameContainer,
     )
 
+    DataContainerRegistry.register_container(
+        LazyType("tritonclient.grpc", "InferInput"),
+        LazyType("tritonclient.grpc", "InferInput"),
+        TritonInferInputDataContainer,
+    )
+    DataContainerRegistry.register_container(
+        LazyType("tritonclient.grpc.aio", "InferInput"),
+        LazyType("tritonclient.grpc.aio", "InferInput"),
+        TritonInferInputDataContainer,
+    )
+    DataContainerRegistry.register_container(
+        LazyType("tritonclient.http", "InferInput"),
+        LazyType("tritonclient.http", "InferInput"),
+        TritonInferInputDataContainer,
+    )
+    DataContainerRegistry.register_container(
+        LazyType("tritonclient.http.aio", "InferInput"),
+        LazyType("tritonclient.http.aio", "InferInput"),
+        TritonInferInputDataContainer,
+    )
+
 
 register_builtin_containers()
 
@@ -444,6 +627,45 @@ class AutoContainer(DataContainer[t.Any, t.Any]):
         container_cls = DataContainerRegistry.find_by_name(payload.container)
         return container_cls.from_payload(payload)
 
+    @t.overload
+    @classmethod
+    def to_triton_payload(
+        cls,
+        inp: tritongrpcclient.InferInput,
+        meta: tritongrpcclient.service_pb2.ModelMetadataResponse.TensorMetadata,
+        _use_http_client: t.Literal[False] = ...,
+    ) -> tritongrpcclient.InferInput:
+        ...
+
+    @t.overload
+    @classmethod
+    def to_triton_payload(
+        cls,
+        inp: tritonhttpclient.InferInput,
+        meta: dict[str, t.Any],
+        _use_http_client: t.Literal[True] = ...,
+    ) -> tritongrpcclient.InferInput:
+        ...
+
+    @classmethod
+    def to_triton_payload(
+        cls, inp: t.Any, meta: t.Any, _use_http_client: bool = False
+    ) -> t.Any:
+        """
+        Convert given input types to a Triton payload.
+
+        Implementation of this method should always return a InferInput that has
+        data of a Numpy array.
+        """
+        if _use_http_client:
+            return DataContainerRegistry.find_by_single_type(
+                type(inp)
+            ).to_triton_http_payload(inp, meta)
+        else:
+            return DataContainerRegistry.find_by_single_type(
+                type(inp)
+            ).to_triton_grpc_payload(inp, meta)
+
     @classmethod
     def batches_to_batch(
         cls, batches: t.Sequence[BatchType], batch_dim: int = 0
