@@ -8,7 +8,6 @@ import typing as t
 import logging
 import tempfile
 import contextlib
-from typing import TYPE_CHECKING
 from pathlib import Path
 from functools import partial
 
@@ -16,10 +15,13 @@ import psutil
 from simple_di import inject
 from simple_di import Provide
 
+from .exceptions import BentoMLException
+from .grpc.utils import LATEST_PROTOCOL_VERSION
 from ._internal.utils import experimental
+from ._internal.runner.runner import Runner
 from ._internal.configuration.containers import BentoMLContainer
 
-if TYPE_CHECKING:
+if t.TYPE_CHECKING:
     from circus.watcher import Watcher
 
 
@@ -91,6 +93,7 @@ def create_watcher(
     name: str,
     args: list[str],
     *,
+    cmd: str = sys.executable,
     use_sockets: bool = True,
     **kwargs: t.Any,
 ) -> Watcher:
@@ -98,7 +101,7 @@ def create_watcher(
 
     return Watcher(
         name=name,
-        cmd=sys.executable,
+        cmd=cmd,
         args=args,
         copy_env=True,
         stop_children=True,
@@ -122,13 +125,7 @@ def log_grpcui_instruction(port: int) -> None:
         platform_deps="host.docker.internal", network_args="-p 8080:8080"
     )
 
-    if os.path.exists("/.dockerenv"):
-        logger.info(
-            "Detected running Bento inside an OCI container. In order to use gRPC UI, do as follows: If your local machine are either MacOS or Windows , then use '%s'. Otherwise use '%s'.",
-            mac_win_instruction,
-            linux_instruction,
-        )
-    elif psutil.WINDOWS or psutil.MACOS:
+    if psutil.WINDOWS or psutil.MACOS:
         logger.info(message, mac_win_instruction)
     elif psutil.LINUX:
         logger.info(message, linux_instruction)
@@ -165,6 +162,15 @@ def construct_ssl_args(
     return args
 
 
+def find_triton_binary():
+    binary = shutil.which("tritonserver")
+    if binary is None:
+        raise RuntimeError(
+            "'tritonserver' is not found on PATH. Make sure to include the compiled binary in PATH to proceed.\nIf you are running this inside a container, make sure to use the official Triton container image as a 'base_image'. See https://catalog.ngc.nvidia.com/orgs/nvidia/containers/tritonserver."
+        )
+    return binary
+
+
 @inject
 def serve_http_development(
     bento_identifier: str,
@@ -186,8 +192,7 @@ def serve_http_development(
 
     from circus.sockets import CircusSocket
 
-    from bentoml import load
-
+    from . import load
     from ._internal.log import SERVER_LOGGING_CONFIG
     from ._internal.utils.circus import create_standalone_arbiter
     from ._internal.utils.analytics import track_serve
@@ -230,7 +235,7 @@ def serve_http_development(
             close_child_stdin=False,
         )
     )
-    scheme = "https" if len(ssl_args) > 0 else "http"
+    scheme = "https" if BentoMLContainer.ssl.enabled.get() else "http"
     if BentoMLContainer.api_server_config.metrics.enabled.get():
         log_host = "localhost" if host == "0.0.0.0" else host
 
@@ -306,12 +311,12 @@ def serve_http_production(
 
     from circus.sockets import CircusSocket
 
-    from bentoml import load
-
+    from . import load
     from ._internal.utils import reserve_free_port
     from ._internal.utils.uri import path_to_uri
     from ._internal.utils.circus import create_standalone_arbiter
     from ._internal.utils.analytics import track_serve
+    from ._internal.configuration.containers import BentoMLContainer
 
     working_dir = os.path.realpath(os.path.expanduser(working_dir))
     svc = load(bento_identifier, working_dir=working_dir, standalone_load=True)
@@ -324,53 +329,14 @@ def serve_http_production(
         # use AF_UNIX sockets for Circus
         uds_path = tempfile.mkdtemp()
         for runner in svc.runners:
-            sockets_path = os.path.join(uds_path, f"{id(runner)}.sock")
-            assert len(sockets_path) < MAX_AF_UNIX_PATH_LENGTH
+            if isinstance(runner, Runner):
+                sockets_path = os.path.join(uds_path, f"{id(runner)}.sock")
+                assert len(sockets_path) < MAX_AF_UNIX_PATH_LENGTH
 
-            runner_bind_map[runner.name] = path_to_uri(sockets_path)
-            circus_socket_map[runner.name] = CircusSocket(
-                name=runner.name,
-                path=sockets_path,
-                backlog=backlog,
-            )
-
-            watchers.append(
-                create_watcher(
-                    name=f"runner_{runner.name}",
-                    args=[
-                        "-m",
-                        SCRIPT_RUNNER,
-                        bento_identifier,
-                        "--runner-name",
-                        runner.name,
-                        "--fd",
-                        f"$(circus.sockets.{runner.name})",
-                        "--working-dir",
-                        working_dir,
-                        "--worker-id",
-                        "$(CIRCUS.WID)",
-                        "--worker-env-map",
-                        json.dumps(runner.scheduled_worker_env_map),
-                        "--prometheus-dir",
-                        prometheus_dir,
-                    ],
-                    working_dir=working_dir,
-                    numprocesses=runner.scheduled_worker_count,
-                )
-            )
-
-    elif psutil.WINDOWS:
-        # Windows doesn't (fully) support AF_UNIX sockets
-        with contextlib.ExitStack() as port_stack:
-            for runner in svc.runners:
-                runner_port = port_stack.enter_context(reserve_free_port())
-                runner_host = "127.0.0.1"
-
-                runner_bind_map[runner.name] = f"tcp://{runner_host}:{runner_port}"
+                runner_bind_map[runner.name] = path_to_uri(sockets_path)
                 circus_socket_map[runner.name] = CircusSocket(
                     name=runner.name,
-                    host=runner_host,
-                    port=runner_port,
+                    path=sockets_path,
                     backlog=backlog,
                 )
 
@@ -391,14 +357,91 @@ def serve_http_production(
                             "$(CIRCUS.WID)",
                             "--worker-env-map",
                             json.dumps(runner.scheduled_worker_env_map),
+                            "--prometheus-dir",
+                            prometheus_dir,
                         ],
                         working_dir=working_dir,
                         numprocesses=runner.scheduled_worker_count,
                     )
                 )
-            port_stack.enter_context(
-                reserve_free_port()
-            )  # reserve one more to avoid conflicts
+            else:
+                # Make sure that the tritonserver uses the correct protocol
+                runner_bind_map[runner.name] = runner.protocol_address
+                cli_args = runner.cli_args + [
+                    f"--http-port={runner.protocol_address.split(':')[-1]}"
+                    if runner.tritonserver_type == "http"
+                    else f"--grpc-port={runner.protocol_address.split(':')[-1]}"
+                ]
+                watchers.append(
+                    create_watcher(
+                        name=f"tritonserver_{runner.name}",
+                        cmd=find_triton_binary(),
+                        args=cli_args,
+                        use_sockets=False,
+                        working_dir=working_dir,
+                        numprocesses=1,
+                    )
+                )
+
+    elif psutil.WINDOWS:
+        # Windows doesn't (fully) support AF_UNIX sockets
+        with contextlib.ExitStack() as port_stack:
+            for runner in svc.runners:
+                if isinstance(runner, Runner):
+                    runner_port = port_stack.enter_context(reserve_free_port())
+                    runner_host = "127.0.0.1"
+
+                    runner_bind_map[runner.name] = f"tcp://{runner_host}:{runner_port}"
+                    circus_socket_map[runner.name] = CircusSocket(
+                        name=runner.name,
+                        host=runner_host,
+                        port=runner_port,
+                        backlog=backlog,
+                    )
+
+                    watchers.append(
+                        create_watcher(
+                            name=f"runner_{runner.name}",
+                            args=[
+                                "-m",
+                                SCRIPT_RUNNER,
+                                bento_identifier,
+                                "--runner-name",
+                                runner.name,
+                                "--fd",
+                                f"$(circus.sockets.{runner.name})",
+                                "--working-dir",
+                                working_dir,
+                                "--worker-id",
+                                "$(CIRCUS.WID)",
+                                "--worker-env-map",
+                                json.dumps(runner.scheduled_worker_env_map),
+                            ],
+                            working_dir=working_dir,
+                            numprocesses=runner.scheduled_worker_count,
+                        )
+                    )
+                else:
+                    # Make sure that the tritonserver uses the correct protocol
+                    runner_bind_map[runner.name] = runner.protocol_address
+                    cli_args = runner.cli_args + [
+                        f"--http-port={runner.protocol_address.split(':')[-1]}"
+                        if runner.tritonserver_type == "http"
+                        else f"--grpc-port={runner.protocol_address.split(':')[-1]}"
+                    ]
+                    watchers.append(
+                        create_watcher(
+                            name=f"tritonserver_{runner.name}",
+                            cmd=find_triton_binary(),
+                            args=cli_args,
+                            use_sockets=False,
+                            working_dir=working_dir,
+                            numprocesses=1,
+                        )
+                    )
+
+            # reserve one more to avoid conflicts
+            port_stack.enter_context(reserve_free_port())
     else:
         raise NotImplementedError("Unsupported platform: {}".format(sys.platform))
 
@@ -420,7 +463,7 @@ def serve_http_production(
         ssl_ca_certs=ssl_ca_certs,
         ssl_ciphers=ssl_ciphers,
     )
-    scheme = "https" if len(ssl_args) > 0 else "http"
+    scheme = "https" if BentoMLContainer.ssl.enabled.get() else "http"
     watchers.append(
         create_watcher(
             name="api_server",
@@ -496,13 +539,13 @@ def serve_grpc_development(
     reload: bool = False,
     channelz: bool = Provide[BentoMLContainer.grpc.channelz.enabled],
     reflection: bool = Provide[BentoMLContainer.grpc.reflection.enabled],
+    protocol_version: str = LATEST_PROTOCOL_VERSION,
 ) -> None:
     prometheus_dir = ensure_prometheus_dir()
 
     from circus.sockets import CircusSocket
 
-    from bentoml import load
-
+    from . import load
     from ._internal.log import SERVER_LOGGING_CONFIG
     from ._internal.utils import reserve_free_port
     from ._internal.utils.circus import create_standalone_arbiter
@@ -525,7 +568,8 @@ def serve_grpc_development(
         ssl_keyfile=ssl_keyfile,
         ssl_ca_certs=ssl_ca_certs,
     )
-    scheme = "https" if len(ssl_args) > 0 else "http"
+
+    scheme = "https" if BentoMLContainer.ssl.enabled.get() else "http"
 
     with contextlib.ExitStack() as port_stack:
         api_port = port_stack.enter_context(
@@ -545,6 +589,8 @@ def serve_grpc_development(
             "--prometheus-dir",
             prometheus_dir,
             *ssl_args,
+            "--protocol-version",
+            protocol_version,
         ]
 
         if reflection:
@@ -674,12 +720,11 @@ def serve_grpc_production(
     | None = Provide[BentoMLContainer.grpc.max_concurrent_streams],
     channelz: bool = Provide[BentoMLContainer.grpc.channelz.enabled],
     reflection: bool = Provide[BentoMLContainer.grpc.reflection.enabled],
+    protocol_version: str = LATEST_PROTOCOL_VERSION,
 ) -> None:
     prometheus_dir = ensure_prometheus_dir()
 
-    from bentoml import load
-    from bentoml.exceptions import UnprocessableEntity
-
+    from . import load
     from ._internal.utils import reserve_free_port
     from ._internal.utils.uri import path_to_uri
     from ._internal.utils.circus import create_standalone_arbiter
@@ -698,7 +743,7 @@ def serve_grpc_production(
     # Check whether users are running --grpc on windows
     # also raising warning if users running on MacOS or FreeBSD
     if psutil.WINDOWS:
-        raise UnprocessableEntity(
+        raise BentoMLException(
             "'grpc' is not supported on Windows with '--production'. The reason being SO_REUSEPORT socket option is only available on UNIX system, and gRPC implementation depends on this behaviour."
         )
     if psutil.MACOS or psutil.FREEBSD:
@@ -707,55 +752,21 @@ def serve_grpc_production(
             "MacOS" if psutil.MACOS else "FreeBSD",
         )
 
+    # NOTE: We need to find and set model-repository args
+    # to all TritonRunner instances (required from tritonserver if spawning multiple instances.)
+
     if psutil.POSIX:
         # use AF_UNIX sockets for Circus
         uds_path = tempfile.mkdtemp()
         for runner in svc.runners:
-            sockets_path = os.path.join(uds_path, f"{id(runner)}.sock")
-            assert len(sockets_path) < MAX_AF_UNIX_PATH_LENGTH
+            if isinstance(runner, Runner):
+                sockets_path = os.path.join(uds_path, f"{id(runner)}.sock")
+                assert len(sockets_path) < MAX_AF_UNIX_PATH_LENGTH
 
-            runner_bind_map[runner.name] = path_to_uri(sockets_path)
-            circus_socket_map[runner.name] = CircusSocket(
-                name=runner.name,
-                path=sockets_path,
-                backlog=backlog,
-            )
-
-            watchers.append(
-                create_watcher(
-                    name=f"runner_{runner.name}",
-                    args=[
-                        "-m",
-                        SCRIPT_RUNNER,
-                        bento_identifier,
-                        "--runner-name",
-                        runner.name,
-                        "--fd",
-                        f"$(circus.sockets.{runner.name})",
-                        "--working-dir",
-                        working_dir,
-                        "--worker-id",
-                        "$(CIRCUS.WID)",
-                        "--worker-env-map",
-                        json.dumps(runner.scheduled_worker_env_map),
-                    ],
-                    working_dir=working_dir,
-                    numprocesses=runner.scheduled_worker_count,
-                )
-            )
-
-    elif psutil.WINDOWS:
-        # Windows doesn't (fully) support AF_UNIX sockets
-        with contextlib.ExitStack() as port_stack:
-            for runner in svc.runners:
-                runner_port = port_stack.enter_context(reserve_free_port())
-                runner_host = "127.0.0.1"
-
-                runner_bind_map[runner.name] = f"tcp://{runner_host}:{runner_port}"
+                runner_bind_map[runner.name] = path_to_uri(sockets_path)
                 circus_socket_map[runner.name] = CircusSocket(
                     name=runner.name,
-                    host=runner_host,
-                    port=runner_port,
+                    path=sockets_path,
                     backlog=backlog,
                 )
 
@@ -772,30 +783,106 @@ def serve_grpc_production(
                             f"$(circus.sockets.{runner.name})",
                             "--working-dir",
                             working_dir,
-                            "--no-access-log",
                             "--worker-id",
-                            "$(circus.wid)",
+                            "$(CIRCUS.WID)",
                             "--worker-env-map",
                             json.dumps(runner.scheduled_worker_env_map),
-                            "--prometheus-dir",
-                            prometheus_dir,
                         ],
                         working_dir=working_dir,
                         numprocesses=runner.scheduled_worker_count,
                     )
                 )
+            else:
+                # Make sure that the tritonserver uses the correct protocol
+                runner_bind_map[runner.name] = runner.protocol_address
+                cli_args = runner.cli_args + [
+                    f"--http-port={runner.protocol_address.split(':')[-1]}"
+                    if runner.tritonserver_type == "http"
+                    else f"--grpc-port={runner.protocol_address.split(':')[-1]}"
+                ]
+                watchers.append(
+                    create_watcher(
+                        name=f"tritonserver_{runner.name}",
+                        cmd=find_triton_binary(),
+                        args=cli_args,
+                        use_sockets=False,
+                        working_dir=working_dir,
+                        numprocesses=1,
+                    )
+                )
+
+    elif psutil.WINDOWS:
+        # Windows doesn't (fully) support AF_UNIX sockets
+        with contextlib.ExitStack() as port_stack:
+            for runner in svc.runners:
+                if isinstance(runner, Runner):
+                    runner_port = port_stack.enter_context(reserve_free_port())
+                    runner_host = "127.0.0.1"
+
+                    runner_bind_map[runner.name] = f"tcp://{runner_host}:{runner_port}"
+                    circus_socket_map[runner.name] = CircusSocket(
+                        name=runner.name,
+                        host=runner_host,
+                        port=runner_port,
+                        backlog=backlog,
+                    )
+
+                    watchers.append(
+                        create_watcher(
+                            name=f"runner_{runner.name}",
+                            args=[
+                                "-m",
+                                SCRIPT_RUNNER,
+                                bento_identifier,
+                                "--runner-name",
+                                runner.name,
+                                "--fd",
+                                f"$(circus.sockets.{runner.name})",
+                                "--working-dir",
+                                working_dir,
+                                "--no-access-log",
+                                "--worker-id",
+                                "$(circus.wid)",
+                                "--worker-env-map",
+                                json.dumps(runner.scheduled_worker_env_map),
+                                "--prometheus-dir",
+                                prometheus_dir,
+                            ],
+                            working_dir=working_dir,
+                            numprocesses=runner.scheduled_worker_count,
+                        )
+                    )
+                else:
+                    # Make sure that the tritonserver uses the correct protocol
+                    runner_bind_map[runner.name] = runner.protocol_address
+                    cli_args = runner.cli_args + [
+                        f"--http-port={runner.protocol_address.split(':')[-1]}"
+                        if runner.tritonserver_type == "http"
+                        else f"--grpc-port={runner.protocol_address.split(':')[-1]}"
+                    ]
+                    watchers.append(
+                        create_watcher(
+                            name=f"tritonserver_{runner.name}",
+                            cmd=find_triton_binary(),
+                            args=cli_args,
+                            use_sockets=False,
+                            working_dir=working_dir,
+                            numprocesses=1,
+                        )
+                    )
             # reserve one more to avoid conflicts
             port_stack.enter_context(reserve_free_port())
     else:
         raise NotImplementedError("Unsupported platform: {}".format(sys.platform))
 
     logger.debug("Runner map: %s", runner_bind_map)
+
     ssl_args = construct_ssl_args(
         ssl_certfile=ssl_certfile,
         ssl_keyfile=ssl_keyfile,
         ssl_ca_certs=ssl_ca_certs,
     )
-    scheme = "https" if len(ssl_args) > 0 else "http"
+    scheme = "https" if BentoMLContainer.ssl.enabled.get() else "http"
 
     with contextlib.ExitStack() as port_stack:
         api_port = port_stack.enter_context(
@@ -818,6 +905,8 @@ def serve_grpc_production(
             "--worker-id",
             "$(CIRCUS.WID)",
             *ssl_args,
+            "--protocol-version",
+            protocol_version,
         ]
         if reflection:
             args.append("--enable-reflection")
@@ -841,7 +930,6 @@ def serve_grpc_production(
             )
         )
 
-    scheme = "https" if len(ssl_args) > 0 else "http"
     if BentoMLContainer.api_server_config.metrics.enabled.get():
         metrics_host = BentoMLContainer.grpc.metrics.host.get()
         metrics_port = BentoMLContainer.grpc.metrics.port.get()
