@@ -25,6 +25,8 @@ from bentoml._internal.utils import LazyLoader
 from bentoml._internal.utils import reserve_free_port
 from bentoml._internal.utils import cached_contextmanager
 
+from ..grpc.utils import LATEST_PROTOCOL_VERSION
+
 if TYPE_CHECKING:
     from grpc import aio
     from grpc_health.v1 import health_pb2 as pb_health
@@ -75,6 +77,7 @@ async def server_warmup(
     check_interval: float = 1,
     popen: subprocess.Popen[t.Any] | None = None,
     service_name: str | None = None,
+    protocol_version: str = LATEST_PROTOCOL_VERSION,
 ) -> bool:
     start_time = time.time()
     proxy_handler = urllib.request.ProxyHandler({})
@@ -86,7 +89,7 @@ async def server_warmup(
 
             try:
                 if service_name is None:
-                    service_name = "bentoml.grpc.v1alpha1.BentoService"
+                    service_name = f"bentoml.grpc.{protocol_version}.BentoService"
                 async with create_channel(host_url) as channel:
                     Check = channel.unary_unary(
                         "/grpc.health.v1.Health/Check",
@@ -127,9 +130,7 @@ async def server_warmup(
 
 
 @cached_contextmanager("{project_path}, {cleanup}")
-def bentoml_build(
-    project_path: str, cleanup: bool = True
-) -> t.Generator[Bento, None, None]:
+def build(project_path: str, cleanup: bool = True) -> t.Generator[Bento, None, None]:
     """
     Build a BentoML project.
     """
@@ -144,62 +145,74 @@ def bentoml_build(
 
 
 @cached_contextmanager("{bento_tag}, {image_tag}, {cleanup}, {use_grpc}")
-def bentoml_containerize(
+def containerize(
     bento_tag: str | Tag,
     image_tag: str | None = None,
     cleanup: bool = True,
     use_grpc: bool = False,
+    backend: str = "docker",
+    **attrs: t.Any,
 ) -> t.Generator[str, None, None]:
     """
     Build the docker image from a saved bento, yield the docker image tag
     """
-    from bentoml import bentos
+    from bentoml import container
 
     bento_tag = Tag.from_taglike(bento_tag)
     if image_tag is None:
-        image_tag = bento_tag.name
+        image_tag = str(bento_tag)
     try:
-        print(f"Building bento server docker image: {bento_tag}")
-        bentos.containerize(
+        print(f"Building bento server container: {bento_tag}")
+        container.build(
             str(bento_tag),
-            docker_image_tag=(image_tag,),
+            backend=backend,
+            image_tag=(image_tag,),
             progress="plain",
             features=["grpc", "grpc-reflection"] if use_grpc else None,
+            label={"testing": True, "module": __name__},
+            **attrs,
         )
         yield image_tag
     finally:
         if cleanup:
-            print(f"Removing bento server docker image: {image_tag}")
-            subprocess.call(["docker", "rmi", image_tag])
+            print(f"Removing bento server container: {image_tag}")
+            subprocess.call([backend, "rmi", image_tag])
 
 
-@cached_contextmanager("{image_tag}, {config_file}, {use_grpc}")
-def run_bento_server_docker(
+@cached_contextmanager(
+    "{image_tag}, {config_file}, {use_grpc}, {protocol_version}, {platform}"
+)
+def run_bento_server_container(
     image_tag: str,
     config_file: str | None = None,
     use_grpc: bool = False,
     timeout: float = 90,
     host: str = "127.0.0.1",
+    backend: str = "docker",
+    protocol_version: str = LATEST_PROTOCOL_VERSION,
+    platform: str = "linux/amd64",
 ):
     """
-    Launch a bentoml service container from a docker image, yield the host URL
+    Launch a bentoml service container from a container, yield the host URL
     """
     from bentoml._internal.configuration.containers import BentoMLContainer
 
-    container_name = f"bentoml-test-{image_tag}-{hash(config_file)}"
+    container_name = f"bentoml-test-{image_tag.replace(':', '_')}-{hash(config_file)}"
     with reserve_free_port(enable_so_reuseport=use_grpc) as port, reserve_free_port(
         enable_so_reuseport=use_grpc
     ) as prom_port:
         pass
 
     cmd = [
-        "docker",
+        backend,
         "run",
         "--rm",
         "--name",
         container_name,
         "--publish",
         f"{port}:3000",
+        "--platform",
+        platform,
     ]
     if config_file is not None:
         cmd.extend(["--env", "BENTOML_CONFIG=/home/bentoml/bentoml_config.yml"])
@@ -213,7 +226,7 @@ def run_bento_server_docker(
     cmd.append(image_tag)
     serve_cmd = "serve-grpc" if use_grpc else "serve-http"
     cmd.extend([serve_cmd, "--production"])
-    print(f"Running API server docker image: '{' '.join(cmd)}'")
+    print(f"Running API server in container: '{' '.join(cmd)}'")
     with subprocess.Popen(
         cmd,
         stdin=subprocess.PIPE,
@@ -222,7 +235,13 @@ def run_bento_server_docker(
         try:
             host_url = f"{host}:{port}"
             if asyncio.run(
-                server_warmup(host_url, timeout=timeout, popen=proc, grpc=use_grpc)
+                server_warmup(
+                    host_url,
+                    timeout=timeout,
+                    popen=proc,
+                    grpc=use_grpc,
+                    protocol_version=protocol_version,
+                )
             ):
                 yield host_url
             else:
@@ -231,7 +250,7 @@ def run_bento_server_docker(
                 ) from None
         finally:
             print(f"Stopping Bento container {container_name}...")
-            subprocess.call(["docker", "stop", container_name])
+            subprocess.call([backend, "stop", container_name])
     time.sleep(1)
 
 
@@ -242,6 +261,7 @@ def run_bento_server_standalone(
     config_file: str | None = None,
     timeout: float = 90,
     host: str = "127.0.0.1",
+    protocol_version: str = LATEST_PROTOCOL_VERSION,
 ):
     """
     Launch a bentoml service directly by the bentoml CLI, yields the host URL.
@@ -272,7 +292,13 @@ def run_bento_server_standalone(
     try:
         host_url = f"{host}:{server_port}"
         assert asyncio.run(
-            server_warmup(host_url, timeout=timeout, popen=p, grpc=use_grpc)
+            server_warmup(
+                host_url,
+                timeout=timeout,
+                popen=p,
+                grpc=use_grpc,
+                protocol_version=protocol_version,
+            )
         )
         yield host_url
     finally:
@@ -297,6 +323,7 @@ def run_bento_server_distributed(
     use_grpc: bool = False,
     timeout: float = 90,
     host: str = "127.0.0.1",
+    protocol_version: str = LATEST_PROTOCOL_VERSION,
 ):
     """
     Launch a bentoml service as a simulated distributed environment(Yatai), yields the host URL.
@@ -386,7 +413,14 @@ def run_bento_server_distributed(
     )
     try:
         host_url = f"{host}:{server_port}"
-        asyncio.run(server_warmup(host_url, timeout=timeout, grpc=use_grpc))
+        asyncio.run(
+            server_warmup(
+                host_url,
+                timeout=timeout,
+                grpc=use_grpc,
+                protocol_version=protocol_version,
+            )
+        )
         yield host_url
     finally:
         for p in processes:
@@ -399,18 +433,21 @@ def run_bento_server_distributed(
 
 
 @cached_contextmanager(
-    "{bento_name}, {project_path}, {config_file}, {deployment_mode}, {bentoml_home}, {use_grpc}"
+    "{bento_name}, {project_path}, {config_file}, {deployment_mode}, {bentoml_home}, {use_grpc}, {protocol_version}"
 )
 def host_bento(
     bento_name: str | Tag | None = None,
     project_path: str = ".",
     config_file: str | None = None,
-    deployment_mode: t.Literal["standalone", "distributed", "docker"] = "standalone",
+    deployment_mode: t.Literal["standalone", "distributed", "container"] = "standalone",
     bentoml_home: str | None = None,
     use_grpc: bool = False,
     clean_context: contextlib.ExitStack | None = None,
     host: str = "127.0.0.1",
     timeout: float = 120,
+    backend: str = "docker",
+    protocol_version: str = LATEST_PROTOCOL_VERSION,
+    container_mode_options: dict[str, t.Any] = None,
 ) -> t.Generator[str, None, None]:
     """
     Host a bentoml service, yields the host URL.
@@ -428,17 +465,24 @@ def host_bento(
         use_grpc: if True, running gRPC tests.
         host: set a given host for the bento, default to ``127.0.0.1``
         timeout: the timeout for the server to start
+        backend: the backend to use for building container, default to ``docker``
 
     Returns:
         :obj:`str`: a generated host URL where we run the test bento.
     """
     import bentoml
 
+    # NOTE: clean_context here to ensure we can close all running context manager
+    # to avoid dangling processes. When running pytest, we have a clean_context fixture
+    # as a ExitStack.
     if clean_context is None:
         clean_context = contextlib.ExitStack()
         clean_on_exit = True
     else:
         clean_on_exit = False
+
+    # NOTE: we need to set the BENTOML_HOME to a temporary folder to avoid
+    # conflict with the user's BENTOML_HOME.
     if bentoml_home:
         from bentoml._internal.configuration.containers import BentoMLContainer
 
@@ -446,7 +490,7 @@ def host_bento(
     try:
         if bento_name is None or not bentoml.list(bento_name):
             bento = clean_context.enter_context(
-                bentoml_build(project_path, cleanup=clean_on_exit)
+                build(project_path, cleanup=clean_on_exit)
             )
         else:
             bento = bentoml.get(bento_name)
@@ -460,18 +504,41 @@ def host_bento(
                 use_grpc=use_grpc,
                 host=host,
                 timeout=timeout,
+                protocol_version=protocol_version,
             ) as host_url:
                 yield host_url
-        elif deployment_mode == "docker":
+        elif deployment_mode == "container":
+            from bentoml._internal.container import REGISTERED_BACKENDS
+
+            if container_mode_options is None:
+                container_mode_options = {}
+
+            if "platform" not in container_mode_options:
+                # NOTE: by default, we use linux/amd64 for the container image
+                container_mode_options["platform"] = "linux/amd64"
+
+            platform = container_mode_options["platform"]
+            if backend not in REGISTERED_BACKENDS:
+                raise ValueError(
+                    f"Unknown backend: {backend}. To register your custom backend, use 'bentoml.container.register_backend()'"
+                )
             container_tag = clean_context.enter_context(
-                bentoml_containerize(bento.tag, use_grpc=use_grpc)
+                containerize(
+                    bento.tag,
+                    use_grpc=use_grpc,
+                    backend=backend,
+                    **container_mode_options,
+                )
             )
-            with run_bento_server_docker(
+            with run_bento_server_container(
                 container_tag,
                 config_file=config_file,
                 use_grpc=use_grpc,
                 host=host,
                 timeout=timeout,
+                backend=backend,
+                protocol_version=protocol_version,
+                platform=platform,
             ) as host_url:
                 yield host_url
         elif deployment_mode == "distributed":
@@ -481,6 +548,7 @@ def host_bento(
                 use_grpc=use_grpc,
                 host=host,
                 timeout=timeout,
+                protocol_version=protocol_version,
             ) as host_url:
                 yield host_url
         else:
