@@ -9,14 +9,13 @@ import bentoml
 from . import FrameworkTestModel
 from . import FrameworkTestModelInput as Input
 from . import FrameworkTestModelConfiguration as Config
+from .torchscript import test_y
+from .torchscript import test_x_list
 
 framework = bentoml.pytorch_lightning
 
 backward_compatible = True
 
-from .torchscript import test_y
-from .torchscript import test_x_list
-
 
 def generate_models():
     class LightningLinearModel(pl.LightningModule):
@@ -25,7 +24,7 @@ def generate_models():
             self.linear = torch.nn.Linear(5, 1, bias=False)
             torch.nn.init.ones_(self.linear.weight)
 
-        def forward(self, x):
+        def forward(self, x: torch.Tensor):  # noqa
             return self.linear(x)
 
     yield LightningLinearModel()
@@ -43,11 +42,11 @@ models = [
                             input_args=[x],
                             expected=lambda out: out == test_y,
                         )
-                        for x in test_x_list
                     ],
                 },
             ),
         ],
     )
     for model in generate_models()
+    for x in test_x_list
 ]
